from musiclang import Melody, Score
from musiclang.transform import ScoreTransformer
from musiclang import Continuation

class MelodyFiller(ScoreTransformer):
    """
    
    """

    def __init__(self, instrument):
        self.instrument = instrument

    def action(self, score: Score, **kwargs):
        pass

        return score


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import numpy as np
import mido
from fractions import Fraction as frac

def voice_to_channel(instrument_list, voice, instruments):
    """

    Parameters
    ----------
    instrument_list :
        
    voice :
        
    instruments :
        

    Returns
    -------

    """
    instrument_of_voice = instruments.get(voice, 0)
    return instrument_list.index(instrument_of_voice)



def matrix_to_events(matrix, 
                     output_file=None, 
                     ticks_per_beat=384,
                     tempo=120, 
                     instruments={}, 
                     time_signature=(4, 4),
                     instrument_names=None):
    matrix = np.asarray(matrix)
    events = [] # [(time, instrument, duration, message)]
    for el in matrix:
        pitch, offset, duration, velocity, track, silence, continuation, tempo, pedal = el
        # Add note on, note off for each track

    return events

def matrix_to_mid(matrix, output_file=None, ticks_per_beat=480, tempo=120, instruments={}, time_signature=(4, 4),
                  anachrusis_time=0,
                  instrument_names=None, one_track_per_instrument=True, **kwargs):
    """

    Parameters
    ----------
    matrix :
        
    output_file :
         (Default value = None)
    ticks_per_beat :
         (Default value = 480)
    tempo :
         (Default value = 120)
    instruments :
         (Default value = {})
    **kwargs :
        

    Returns
    -------

    """
    if instrument_names is None:
        instrument_names = []
    from mido import MidiFile, MidiTrack, Message, MetaMessage
    from ..constants import SILENCE, CONTINUATION, PITCH, TRACK, OFFSET, VELOCITY, DURATION, TEMPO, PEDAL
    from .constants import OCTAVES
    import os

    matrix = np.asarray(matrix)

    #


    if one_track_per_instrument:
        instrument_names_new = []
        instruments_base = dict(instruments)
        instrument_group = {}
        for track_nb, program in instruments.items():
            instrument_name = instrument_names[track_nb]
            if instrument_name.startswith('drums_'):
                instrument_group[-1] = instrument_group.get(-1, []) + [track_nb]
            else:
                instrument_group[program] = instrument_group.get(program, []) + [track_nb]

        # Change the tracks associated to each instrument
        instruments = {i: program for i, program in enumerate(instrument_group.keys())}
        invert_instruments = {program: i for i, program in instruments.items()}
        instruments = {key: val if val != -1 else 0 for key, val in instruments.items() }
        # Associate each track to the right new track
        for program, tracks in instrument_group.items():
            instrument_name = instrument_names[tracks[0]]
            if program == -1:
                instrument_name = 'drums_0'
            instrument_names_new.append(instrument_name)
            for i, track in enumerate(tracks):
                matrix[matrix[:, TRACK] == track, TRACK] = invert_instruments[program]

        instrument_names = instrument_names_new

    def number_to_channel(n):
        """

        Parameters
        ----------
        n :
            

        Returns
        -------

        """
        if n < 9:
            return n
        if n >= 9:
            return n + 1

    mid = MidiFile()
    mid.ticks_per_beat = ticks_per_beat
    mid.type = 1

    bar_duration = time_signature[0] * 4 / time_signature[1]
    start_time = (-anachrusis_time) % bar_duration

    from fractions import Fraction as frac
    matrix[:, OFFSET] = matrix[:, OFFSET] + frac(start_time).limit_denominator(24)
    # Take care of continuation
    # matrix = matrix[(matrix['pitch'] > 0) | (matrix['continuation'] > 0)]
    # Remove all continuations that follows a silence
    # Put all instruments with same name in same track
    # For each voice assign right track
    instrument_list = list(sorted(list(set([0] + list(instruments.values())))))
    # Replace with track in matrix
    # For each voice assign right track

    # Replace with track in matrix
    nb_tracks = matrix[:, TRACK].max()
    channels = [number_to_channel(voice_to_channel(instrument_list, i, instruments)) for i in range(nb_tracks + 1)]

    for i in range(nb_tracks + 1):
        track = MidiTrack()
        if (len(instrument_names) > i) and instrument_names[i].startswith('drum'):
            channels[i] = 9
            track.append(Message('program_change', program=0, time=0, channel=9))
        elif i in instruments.keys():
            track.append(Message('program_change', program=instruments.get(i, 0), time=0, channel=channels[i]))
        elif i != 9:
            track.append(Message('program_change', program=0, time=0, channel=channels[i]))
        mid.tracks.append(track)

    if isinstance(output_file, str):
        mid.tracks[0].append(MetaMessage("track_name", name=os.path.split(output_file)[1], time=int(0)))
    else:
        mid.tracks[0].append(MetaMessage("track_name", name='track', time=int(0)))
    mid.tracks[0].append(MetaMessage("set_tempo", tempo=mido.bpm2tempo(tempo), time=int(0)))
    mid.tracks[0].append(MetaMessage("time_signature", numerator=time_signature[0], denominator=time_signature[1], time=int(0)))



    sort_events = []
    last_silence = True
    for row in matrix:
        pitch = row[PITCH] + 60 + 12 * OCTAVES.get(instruments.get(row[TRACK], 0), 0)
        row = row.tolist()
        if row[SILENCE]:
            last_silence = True
            sort_events.append([0, 2, row[OFFSET], row[TEMPO], row[PEDAL], row[TRACK]])
            continue

        if row[CONTINUATION]:
            sort_events.append([0, 2, row[OFFSET], row[TEMPO], row[PEDAL], row[TRACK]])

        if row[CONTINUATION] == 0:
            sort_events.append([pitch,
                                1, row[OFFSET], row[VELOCITY], row[TRACK], row[TEMPO], row[PEDAL]])

        if not (row[CONTINUATION] and last_silence):
            sort_events.append([pitch, 0, (row[OFFSET] + row[DURATION]), row[TRACK], row[CONTINUATION],
                                row[TEMPO], row[PEDAL]
                                ])
            last_silence = False

    sort_events.sort(key=lambda tup: tup[2])
    from fractions import Fraction as frac
    current_tempo = tempo
    current_pedal = None
    lapso = {idx: 0 for idx in range(nb_tracks + 1)}
    prev_pitch = {idx: 0 for idx in range(nb_tracks + 1)}
    for evt in sort_events:
        if evt[1] == 1:
            vel = int(evt[3])
            track_nb = int(evt[4])
            track = mid.tracks[track_nb]
            if evt[0] > 100 or evt[0] < 10:
                evt[0] = 0
                vel = 0

            real_time = int((evt[2] - lapso[track_nb]) * ticks_per_beat)

            track.append(Message('note_on', note=int(evt[0]), channel=channels[track_nb],
                                 velocity=vel, time=real_time))

            if evt[5] is not None and evt[5] != current_tempo:
                real_tempo =(480000 * 120) // evt[5]
                track.append(MetaMessage("set_tempo", tempo=real_tempo, time=int(0)))
                current_tempo = evt[5]

            if evt[6] is not None and evt[6]:
                track.append(Message('control_change', value=127, channel=channels[track_nb], control=4, time=int(0)))

            if evt[6] is not None and not evt[6]:
                track.append(Message('control_change', value=0, channel=channels[track_nb], control=4, time=int(0)))

            lapso[track_nb] = evt[2]
            prev_pitch[track_nb] = int(evt[0])
        elif evt[1] == 0:
            track_nb = int(evt[3])
            track = mid.tracks[track_nb]
            if evt[0] > 100 or evt[0] < 10:
                evt[0] = 0
            if evt[4] == 1 and prev_pitch[track_nb] > 0:
                track[-1] = Message('note_off', note=prev_pitch[track_nb], channel=channels[track_nb],
                                    velocity=0, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat + track[-1].time))


            else:
                track.append(Message('note_off', note=int(evt[0]), channel=channels[track_nb],
                                     velocity=0, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat)))
            lapso[track_nb] = evt[2]

        elif evt[1] == 2:
            track_nb = int(evt[5])
            track = mid.tracks[track_nb]
            tempo_change = evt[3]
            pedal_change = evt[4]

            if tempo_change is not None and tempo_change != current_tempo:
                real_tempo = (480000 * 120) // tempo_change
                track.append(MetaMessage("set_tempo", tempo=real_tempo, time=int(0)))
                current_tempo = tempo_change

            if pedal_change is not None and pedal_change:
                track.append(Message('control_change', value=127, channel=channels[track_nb], control=4, time=int(0)))

            if pedal_change is not None and not pedal_change:
                track.append(Message('control_change', value=0, channel=channels[track_nb], control=4, time=int(0)))


    if output_file is not None:
        for track in mid.tracks:
            track.append(MetaMessage('end_of_track', time=(int(0))))
        if isinstance(output_file, str):
            mid.save(output_file)
        else:
            mid.save(file=output_file)

    return mid



def matrix_to_mid_one_channel_per_track(matrix, output_file=None, ticks_per_beat=96, tempo=120, instruments={}, **kwargs):
    """

    Parameters
    ----------
    matrix :
        
    output_file :
         (Default value = None)
    ticks_per_beat :
         (Default value = 96)
    tempo :
         (Default value = 120)
    instruments :
         (Default value = {})
    **kwargs :
        

    Returns
    -------

    """
    from mido import MidiFile, MidiTrack, Message, MetaMessage
    from ..constants import SILENCE, CONTINUATION, PITCH, TRACK, OFFSET, VELOCITY, DURATION
    from .constants import OCTAVES
    import os

    matrix = np.asarray(matrix)

    def number_to_channel(n):
        """

        Parameters
        ----------
        n :
            

        Returns
        -------

        """
        if n < 9:
            return n
        if n >= 9:
            return n + 1

    mid = MidiFile()
    mid.ticks_per_beat = ticks_per_beat
    mid.type = 1

    # Take care of continuation
    # matrix = matrix[(matrix['pitch'] > 0) | (matrix['continuation'] > 0)]
    # Remove all continuations that follows a silence
    # Put all instruments with same name in same track


    # For each voice assign right track
    instrument_list = list(set(list(instruments.values())))

    # Replace with track in matrix

    nb_tracks = matrix[:, TRACK].max()
    for i in range(nb_tracks + 1):
        track = MidiTrack()
        if i in instruments.keys():
            track.append(Message('program_change', program=instruments.get(i, 0), time=0, channel=number_to_channel(i)))
        elif i != 9:
            track.append(Message('program_change', program=0, time=0, channel=number_to_channel(i)))
        mid.tracks.append(track)

    if output_file is not None:
        mid.tracks[0].append(MetaMessage("track_name", name=os.path.split(output_file)[1], time=int(0)))
        mid.tracks[0].append(MetaMessage("set_tempo", tempo=(480000 * 120) // tempo, time=int(0)))
        mid.tracks[0].append(MetaMessage("time_signature", numerator=4, denominator=4, time=int(0)))

    sort_events = []
    last_silence = True
    for row in matrix:
        pitch = row[PITCH] + 60 + 12 * OCTAVES.get(instruments.get(row[TRACK], 0), 0)
        row = row.tolist()
        if row[SILENCE]:
            last_silence = True
            continue
        if row[CONTINUATION] == 0:
            sort_events.append([pitch,
                                1, row[OFFSET], row[VELOCITY], row[TRACK]])

        if not (row[CONTINUATION] and last_silence):
            sort_events.append([pitch, 0, (row[OFFSET] + row[DURATION]), row[TRACK], row[CONTINUATION]])
            last_silence = False

    sort_events.sort(key=lambda tup: tup[2])
    from fractions import Fraction as frac

    lapso = {idx: 0 for idx in range(nb_tracks + 1)}
    prev_pitch = {idx: 0 for idx in range(nb_tracks + 1)}
    for evt in sort_events:
        if evt[1] == 1:
            vel = int(evt[3])
            track_nb = int(evt[4])
            track = mid.tracks[track_nb]
            track.append(Message('note_on', note=int(evt[0]), channel=number_to_channel(track_nb),
                                 velocity=vel, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat)))

            lapso[track_nb] = evt[2]
            prev_pitch[track_nb] = int(evt[0])
        elif evt[1] == 0:
            track_nb = int(evt[3])
            track = mid.tracks[track_nb]

            if evt[4] == 1 and prev_pitch[track_nb] > 0:
                track[-1] = Message('note_off', note=prev_pitch[track_nb], channel=number_to_channel(track_nb),
                                    velocity=0, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat + track[-1].time))


            else:
                track.append(Message('note_off', note=int(evt[0]), channel=number_to_channel(track_nb),
                                     velocity=0, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat)))
            lapso[track_nb] = evt[2]

    if output_file is not None:
        for track in mid.tracks:
            track.append(MetaMessage('end_of_track', time=(int(0))))
        mid.save(output_file)
    return mid



import glob
import random
from musiclang import Score, Tonality
import numpy as np
from musiclang.library import *

def map_nearest_indices(A, B):
    def nearest_index(a, b):
        return min(range(len(a)), key=lambda i: abs(a[i] - b))

    return {i: nearest_index(A, b) for i, b in enumerate(B)}


def map_matrix(A, B, Mb):
    """
    Map a matrix Mb to a matrix Ma, such that Ma[i] is the row in Mb that is closest to B[i]
    Parameters
    ----------
    A: list
    B: list
    Mb: matrix

    Returns
    -------
    Ma: matrix

    """
    Mb = np.asarray(Mb)
    if not len(B) == Mb.shape[0]:
        raise ValueError("Length of B must be equal to the number of rows in Mb")

    mapping = map_nearest_indices(A, B)
    Ma = np.zeros((len(A), Mb.shape[1]), dtype=int)

    for i in range(len(A)):
        # Finding the indices in B that are mapped to the current index in A
        mapped_indices = [index for index, mapped_index in mapping.items() if mapped_index == i]

        # Averaging the rows in Mb that are mapped to the current row in Ma
        if mapped_indices:
            Ma[i] = np.sum(Mb[mapped_indices], axis=0)

    return Ma.tolist()


def chord_to_orchestra(chord, drop_drums=True, nb=4, pattern=False):
    """
    Convert a chord to an orchestra
    Parameters
    ----------
    chord: Chord
        The musiclang chord to convert into an orchestra
    drop_drums : bool, optional, default=True
        Drop drums instruments
    nb: int, optional, default=4
        nb arpeggio notes to consider on each side, it will give the acceptable range of the instrument melody


    Returns
    -------
    l:list
    Orchestra object, one entry per instrument

    """
    orchestra_raw = chord.get_orchestration()
    orchestra = []

    # chord_tonality =
    for instrument in orchestra_raw:
        if drop_drums and instrument['instrument'].startswith('drums_0'):
            continue
        else:
            note = instrument['note']
            base_note = chord.__getattribute__(note)
            base_pitch = chord.to_pitch(base_note)
            pitch_grid = (
                    [chord.to_pitch(Score.from_str(f'bd{nb - i}'), last_pitch=base_pitch) for i in range(nb)]
                    + [base_pitch]
                    + [chord.to_pitch(Score.from_str(f'bu{i + 1}'), last_pitch=base_pitch) for i in range(nb)]
            )
            notes_grid = (
                    [f'bd{nb - i}' for i in range(nb)]
                    + ['x0']
                    + [f'bu{i + 1}' for i in range(nb)]
            )

            melody_score = instrument['pattern'].apply_pattern(base_note)
            grid = chord(melody_score).to_score().to_melody_grid('piano__0', octave=0)
            pitch_range = grid['notes']

            # Convert absolute notes to arpeggio
            grid['rhythm'] = map_matrix(pitch_grid, pitch_range, grid['rhythm'])
            instrument['rhythm'] = grid
            instrument['rhythm']['notes'] = notes_grid
            instrument['rhythm']['octave'] = instrument['octave']
            if not pattern:
                del instrument['pattern']
            orchestra.append(instrument)

    return orchestra


from .element import Formatter

class FreeTextElement(Formatter):

    def __init__(self, text):
        self.text = text

    def to_text(self):
        return '\n' + str(self)

    def __repr__(self):
        return str(self.text)



"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

ELEMENT_TO_STR = {
    0: "I",
    1: "II",
    2: "III",
    3: "IV",
    4: "V",
    5: "VI",
    6: "VII"
}

DEGREE_TO_STR = {
    0: "I",
    1: "II.b",
    2: "II",
    3: "III.b",
    4: "III",
    5: "IV",
    6: "IV.s",
    7: "V",
    8: "VI.b",
    9: "VI",
    10: "VII.b",
    11: "VII"
}

DEGREE_TO_SCALE_DEGREE = {
    0: 0,
    1: 1,
    2: 1,
    3: 2,
    4: 2,
    5: 3,
    6: 3,
    7: 4,
    8: 5,
    9: 5,
    10: 6,
    11: 6
}

SCALE_DEGREE = {
    0: 0,
    1: 2,
    2: 4,
    3: 5,
    4: 7,
    5: 9,
    6: 11
}

from fractions import Fraction as frac

W = frac(4)
H = frac(2)
Q = frac(1)
E = frac(1) / frac(2)
S = frac(1) / frac(4)
T = frac(1) / frac(8)

W3 = (frac(2) / frac(3)) * W
H3 = (frac(2) / frac(3)) * H
Q3 = (frac(2) / frac(3)) * Q
E3 = (frac(2) / frac(3)) * E
S3 = (frac(2) / frac(3)) * S
T3 = (frac(2) / frac(3)) * T

W5 = (frac(2) / frac(5)) * W
H5 = (frac(2) / frac(5)) * H
Q5 = (frac(2) / frac(5)) * Q
E5 = (frac(2) / frac(5)) * E
S5 = (frac(2) / frac(5)) * S
T5 = (frac(2) / frac(5)) * T

Wd = W * frac(3) / frac(2)
Hd = H * frac(3) / frac(2)
Qd = Q * frac(3) / frac(2)
Ed = E * frac(3) / frac(2)
Sd = S * frac(3) / frac(2)
Td = T * frac(3) / frac(2)

STR_TO_DURATION = {
    "n": 0,
    # We let possibility for empty notes because it can be useful (eg : when projecting a melody), these should be ignored by parser
    "w": frac(4),
    "h": frac(2),
    "q": frac(1),
    "e": frac(1) / frac(2),
    "s": frac(1) / frac(4),
    "t": frac(1) / frac(8),

    "w5": frac(8) / frac(5),
    "h5": frac(4) / frac(5),
    "q5": frac(2) / frac(5),
    "e5": frac(2) / frac(10),
    "s5": frac(2) / frac(20),
    "t5": frac(2) / frac(40),

    "w7": frac(8) / frac(7),
    "h7": frac(4) / frac(7),
    "q7": frac(2) / frac(7),
    "e7": frac(2) / frac(14),
    "s7": frac(2) / frac(28),
    "t7": frac(2) / frac(56),

    "w3": frac(8) / frac(3),
    "h3": frac(4) / frac(3),
    "q3": frac(2) / frac(3),
    "e3": frac(2) / frac(6),
    "s3": frac(2) / frac(12),
    "t3": frac(2) / frac(24),

    "wd": W * frac(3) / frac(2),
    "hd": H * frac(3) / frac(2),
    "qd": Q * frac(3) / frac(2),
    "ed": E * frac(3) / frac(2),
    "sd": S * frac(3) / frac(2),
    "td": T * frac(3) / frac(2),
}

DURATION_TO_STR = {
    val: key for key, val in STR_TO_DURATION.items()
}

INDEX_MODE = {
    "M": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 4,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 8,
        (4, "b"): 6,

        (5, ""): 9,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 11,
        (6, "s"): 11,
        (6, "b"): 10,

    },

    "m": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 3,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 7,
        (4, "b"): 6,

        (5, ""): 8,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 11,
        (6, "s"): 11,
        (6, "b"): 10,
    },

    "mm": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 3,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 7,
        (4, "b"): 6,

        (5, ""): 9,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 11,
        (6, "s"): 11,
        (6, "b"): 10,
    },

    "dorian": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 3,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 7,
        (4, "b"): 6,

        (5, ""): 9,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 10,
        (6, "s"): 11,
        (6, "b"): 10,
    },

    "phrygian": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 1,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 3,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 7,
        (4, "b"): 6,

        (5, ""): 9,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 10,
        (6, "s"): 11,
        (6, "b"): 10,
    },

    "lydian": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 4,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 6,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 8,
        (4, "b"): 6,

        (5, ""): 9,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 11,
        (6, "s"): 11,
        (6, "b"): 10,

    },

    "mixolydian": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 4,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 8,
        (4, "b"): 6,

        (5, ""): 9,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 10,
        (6, "s"): 11,
        (6, "b"): 10,

    },

    "aeolian": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 2,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 3,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 7,
        (4, "s"): 7,
        (4, "b"): 6,

        (5, ""): 8,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 10,
        (6, "s"): 11,
        (6, "b"): 10,
    },

    "locrian": {
        (0, ""): 0,
        (0, "s"): 0,
        (0, "b"): 0,

        (1, ""): 1,
        (1, "s"): 2,
        (1, "b"): 1,

        (2, ""): 3,
        (2, "s"): 4,
        (2, "b"): 3,

        (3, ""): 5,
        (3, "s"): 6,
        (3, "b"): 5,

        (4, ""): 6,
        (4, "s"): 7,
        (4, "b"): 6,

        (5, ""): 8,
        (5, "s"): 9,
        (5, "b"): 8,

        (6, ""): 10,
        (6, "s"): 11,
        (6, "b"): 10,
    },

}


ACCIDENTS_TO_NOTE = {
    (0, "min"): 0,
    (0, "maj"): 0,
    (0, "natural"): 0,
    (0, "dim"): 0,
    (0, "aug"): 1,

    (1, "min"): 1,
    (1, "maj"): 2,
    (1, "natural"): 2,
    (1, "dim"): 1,
    (1, "aug"): 2,

    (2, "min"): 3,
    (2, "maj"): 4,
    (2, "natural"): 4,
    (2, "dim"): 3,
    (2, "aug"): 4,

    (3, "min"): 5,
    (3, "maj"): 5,
    (3, "natural"): 5,
    (3, "dim"): 5,
    (3, "aug"): 6,

    (4, "min"): 7,
    (4, "maj"): 7,
    (4, "natural"): 7,
    (4, "dim"): 6,
    (4, "aug"): 7,

    (5, "min"): 8,
    (5, "maj"): 9,
    (5, "natural"): 9,
    (5, "dim"): 8,
    (5, "aug"): 9,

    (6, "min"): 10,
    (6, "maj"): 11,
    (6, "natural"): 11,
    (6, "dim"): 10,
    (6, "aug"): 11,
}



def construct_index_tonality():
    """ """
    result = {}
    for mode in INDEX_MODE.keys():
        D = INDEX_MODE[mode]
        res_mode = {}
        for key, val in D.items():
            if val not in res_mode:
                res_mode[val] = key
            elif key[1] == "":
                res_mode[val] = key
        result[mode] = res_mode

        # Choose best
    return result


INDEX_TONALITY = construct_index_tonality()

SCALES = {
    "M": [0, 2, 4, 5, 7, 9, 11],
    "m": [0, 2, 3, 5, 7, 8, 11],
    "mm": [0, 2, 3, 5, 7, 9, 11],
    "dorian": [0, 2, 3, 5, 7, 9, 10],
    "phrygian": [0, 1, 3, 5, 7, 9, 10],
    "lydian": [0, 2, 4, 6, 7, 9, 11],
    "mixolydian": [0, 2, 4, 5, 7, 9, 10],
    "aeolian": [0, 2, 3, 5, 7, 8, 10],
    "locrian": [0, 1, 3, 5, 6, 8, 10]
}

SCALE_UNISSON = 0
SCALE_SECOND = 1
SCALE_THIRD = 2
SCALE_FOURTH = 3
SCALE_FIFTH = 4
SCALE_SIXTH = 5
SCALE_SEVENTH = 6
SCALE_OCTAVE = 7

INST_PIANO = 'piano'
INST_PIZZICATO = 'pizzicato'
INST_BRIGHT_PIANO = 'bright_piano'
INST_ELECTRIC_PIANO = 'electric_piano'
INST_HONKY_TONK = 'honky_tonk'
INST_ELECTRIC_PIANO_1 = 'electric_piano_1'
INST_ELECTRIC_PIANO_2 = 'electric_piano_2'
INST_HARPSICHORD = 'harpsichord'
INST_CLAVI = 'clavi'
INST_CELESTA = 'celesta'
INST_GLOCKENSPIEL = 'glockenspiel'
INST_MUSIC_BOX = 'music_box'
INST_VIBRAPHONE = 'vibraphone'
INST_MARIMBA = 'marimba'
INST_XYLOPHONE = 'xylophone'
INST_TUBULAR_BELLS = 'tubular_bells'
INST_DULCIMER = 'dulcimer'
INST_DRAWBAR_ORGAN = 'drawbar_organ'
INST_PERCUSSIVE_ORGAN = 'percussive_organ'
INST_ROCK_ORGAN = 'rock_organ'
INST_CHURCH_ORGAN = 'church_organ'
INST_REED_ORGAN = 'reed_organ'
INST_ACCORDION = 'accordion'
INST_HARMONICA = 'harmonica'
INST_TANGO_ACCORDION = 'tango_accordion'
INST_ACOUSTIC_GUITAR = 'acoustic_guitar'
INST_STEEL_GUITAR = 'steel_guitar'
INST_JAZZ_GUITAR = 'jazz_guitar'
INST_CLEAN_GUITAR = 'clean_guitar'
INST_OVERDRIVEN_GUITAR = 'overdriven_guitar'
INST_DISTORTION_GUITAR = 'distortion_guitar'
INST_HARMONIC_GUITAR = 'harmonic_guitar'
INST_ACOUSTIC_BASS = 'acoustic_bass'
INST_ELECTRIC_BASS_FINGER = 'electric_bass_finger'
INST_ELECTRIC_BASS_PICK = 'electric_bass_pick'
INST_FRETLESS_BASS = 'fretless_bass'
INST_SLAP_BASS_1 = 'slap_bass_1'
INST_SLAP_BASS_2 = 'slap_bass_2'
INST_SYNTH_BASS_1 = 'synth_bass_1'
INST_SYNTH_BASS_2 = 'synth_bass_2'
INST_VIOLIN = 'violin'
INST_VIOLA = 'viola'
INST_CELLO = 'cello'
INST_CONTRABASS = 'contrabass'
INST_TREMOLO_STRING = 'tremolo_string'
INST_HARP = 'harp'
INST_TIMPANI = 'timpani'
INST_STRING_ENSEMBLE_1 = 'string_ensemble_1'
INST_STRING_ENSEMBLE_2 = 'string_ensemble_2'
INST_SYNTH_STRING_1 = 'synth_string_1'
INST_SYNTH_STRING_2 = 'synth_string_2'
INST_CHOIR_AAHS = 'choir_aahs'
INST_CHOIR_OOHS = 'choir_oohs'
INST_SYNTH_CHOIR = 'synth_choir'
INST_ORCHESTRA_HIT = 'orchestra_hit'
INST_TRUMPET = 'trumpet'
INST_TROMBONE = 'trombone'
INST_TUBA = 'tuba'
INST_MUTED_TRUMPET = 'muted_trumpet'
INST_FRENCH_HORN = 'french_horn'
INST_BRASS_SECTION = 'brass_section'
INST_SYNTH_BRASS_1 = 'synth_brass_1'
INST_SYNTH_BRASS_2 = 'synth_brass_2'
INST_SOPRANO_SAX = 'soprano_sax'
INST_ALTO_SAX = 'alto_sax'
INST_TENOR_SAX = 'tenor_sax'
INST_BARITONE_SAX = 'baritone_sax'
INST_OBOE = 'oboe'
INST_ENGLISH_HORN = 'english_horn'
INST_BASSOON = 'bassoon'
INST_CLARINET = 'clarinet'
INST_PICCOLO = 'piccolo'
INST_FLUTE = 'flute'
INST_RECORDER = 'recorder'
INST_PAN_FLUTE = 'pan_flute'
INST_BLOWN_BOTTLE = 'blown_bottle'
INST_SHAKUHACHI = 'shakuhachi'
INST_WHISTLE = 'whistle'
INST_OCARINA = 'ocarina'
INST_SQUARE_LEAD = 'square_lead'
INST_SAWTOOTH_LEAD = 'sawtooth_lead'
INST_CALLIOPE_LEAD = 'calliope_lead'
INST_CHIFF_LEAD = 'chiff_lead'
INST_CHARANG_LEAD = 'charang_lead'
INST_VOICE_LEAD = 'voice_lead'
INST_FIFTHS_LEAD = 'fifths_lead'
INST_BASS_LEAD = 'bass_lead'
INST_PAD_NEW_AGE = 'pad_new_age'
INST_PAD_WARM = 'pad_warm'
INST_PAD_POLYSYNTH = 'pad_polysynth'
INST_PAD_CHOIR = 'pad_choir'
INST_PAD_BOWED = 'pad_bowed'
INST_PAD_METALLIC = 'pad_metallic'
INST_PAD_HALO = 'pad_halo'
INST_PAD_SWEEP = 'pad_sweep'
INST_FX_RAIN = 'fx_rain'
INST_FX_SOUNDTRACK = 'fx_soundtrack'
INST_FX_CRYSTAL = 'fx_crystal'
INST_FX_ATMOSPHERE = 'fx_atmosphere'
INST_FX_BRIGHTNESS = 'fx_brightness'
INST_FX_GLOBLINS = 'fx_globlins'
INST_FX_ECHOES = 'fx_echoes'
INST_FX_SCI_FI = 'fx_sci_fi'
INST_SITAR = 'sitar'
INST_BANJO = 'banjo'
INST_SHAMISEN = 'shamisen'
INST_KOTO = 'koto'
INST_KALIMBA = 'kalimba'
INST_BAGPIPE = 'bagpipe'
INST_FIDDLE = 'fiddle'
INST_SHANAI = 'shanai'
INST_TINKLE_BELL = 'tinkle_bell'
INST_AGOGO = 'agogo'
INST_STEEL_DRUMS = 'steel_drums'
INST_WOODBLOCK = 'woodblock'
INST_TAIKO_DRUM = 'taiko_drum'
INST_MELODIC_TOM = 'melodic_tom'
INST_SYNTH_DRUM = 'synth_drum'
INST_REVERSE_CYMBAL = 'reverse_cymbal'
INST_GUITAR_FRET_NOISE = 'guitar_fret_noise'
INST_BREATH_NOISE = 'breath_noise'
INST_SEASHORE = 'seashore'
INST_BIRD_TWEET = 'bird_tweet'
INST_TELEPHONE_RING = 'telephone_ring'
INST_HELICOPTER = 'helicopter'
INST_APPLAUSE = 'applause'
INST_GUNSHOT = 'gunshot'

OCTAVES = {
    INST_PICCOLO: 1,
    INST_FLUTE: 1,
    INST_OBOE: 0,
    INST_ENGLISH_HORN: 0,
    INST_CLARINET: 0,
    INST_BASSOON: -1,
    INST_FRENCH_HORN: -1,
    INST_TRUMPET: -1,
    INST_TROMBONE: -2,
    INST_TUBA: -2,
    INST_PIANO: -1,
    INST_VIOLIN: 1,
    INST_VIOLA: 0,
    INST_CELLO: -2,
    INST_CONTRABASS: -2,
    INST_HARP: -1,
    INST_ACOUSTIC_GUITAR: -1,
    INST_TIMPANI: -2
}
OCTAVES = {
    key: 0 for key in OCTAVES.keys()
}

ALL_INST = list(OCTAVES.keys())

PITCH = 0
OFFSET = 1
DURATION = 2
VELOCITY = 3
TRACK = 4
SILENCE = 5
CONTINUATION = 6
TEMPO = 7
PEDAL = 8

DRUMS_DICT = {
    (0, -2):  'bd',
    (1, -2): 'rs',
    (2, -2): 'sn',
    (3, -2): 'cp',
    (4, -2): 'sn2',
    (5, -2): 'bt',
    (6, -2): 'hh',
    (7, -2): 'lt',
    (8, -2): 'ch',
    (9, -2): 'mt',
    (10, -2): 'oh',
    (11, -2): 'ht'
}

_NOTES = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Fs', 'G', 'Ab', 'A', 'Bb', 'B']
def absolute_note_repr(note):
    val, oct = note.val, note.octave
    note_str = _NOTES[val % len(_NOTES)]
    oct_str = str(note.base_octave + oct + val // oct)
    return f"{note_str}{oct_str}"


from musiclang import Score, Melody
import numpy as np
from fractions import Fraction as frac
from musiclang.library import *


def grid_to_melody(grid, tatum, notes, mode='legato'):
    """
    A grid in an array of array (dimensions : notes, time)
    Parameters
    ----------
    grid

    Returns
    -------

    """

    notes = [Score.from_str(n) for n in notes]
    step = frac(*tatum)
    grid = np.asarray(grid)

    melody = []
    for j, col in enumerate(grid.T):
        for i, row in enumerate(col):
            if row:
                note = notes[i]
                note.duration = step
                melody.append(x0.n)
                melody.append(note)
                break
        else:

            if mode == 'staccato' or j == 0:
                melody.append(r.set_duration(step))
            elif mode == 'legato':
                melody.append(l.set_duration(step))

    return Melody(melody)



from musiclang import Score, Tonality
import numpy as np
from scipy.spatial.distance import cdist

from .chord_inference_utils import TEMPLATES, COEFFS, CHROMA_VECTORS_MATRIX, PITCH_CLASSES_DICT, CHORD_TYPE_TO_PITCHES, CHORD_TYPES
def max_correlation_index(input_matrix, template_matrix):
    # Calculate correlations
    correlations = 1 - cdist(input_matrix, template_matrix, 'correlation')
    # Multiply the correlations by coefficients
    correlations *= COEFFS
    # Find index of max correlation for each row
    max_corr_indices = np.argmax(correlations, axis=1)
    chord_type = max_corr_indices % len(TEMPLATES)
    chord_root = max_corr_indices // len(TEMPLATES)
    return chord_root, chord_type


def filter_notes(notes, start, end):
    pass

def get_chroma_vectors(notes, bars):
    drums_array = notes[:, 5] == 9
    # Remove drums
    notes = notes[~drums_array]
    times_array = np.asarray([notes[:, 0], notes[:, 0] + notes[:, 2]])
    pitches_class_array = notes[:, 1] % 12
    # Convert pitch class array to a 12xnb_notes matrix with each row being the indexes of the pitch class
    pitches_class_matrix = np.asarray([[1.0 * ((i%12) == pitch_class) for i in range(12)] for pitch_class in pitches_class_array])
    starts = np.asarray([bar[0] for bar in bars])
    ends = np.asarray([bar[1] for bar in bars])
    # Create a nb_notesxnb_bars array with each cell being the intersection length between the note and the bar
    intersections = np.maximum(0, np.minimum(times_array[1, :, None], ends[None, :]) - np.maximum(times_array[0, :, None], starts[None, :])).T
    bar_chroma_vectors = intersections.dot(pitches_class_matrix)
    # Normalize each rows
    try:
        normalizer = bar_chroma_vectors.sum(axis=1)[:, None]
    except:
        bar_chroma_vectors = np.zeros((len(bars), 12))
        normalizer = bar_chroma_vectors.sum(axis=1)[:, None]

    normalizer[normalizer == 0] = 1
    bar_chroma_vectors = bar_chroma_vectors / normalizer
    return bar_chroma_vectors


def get_bass_note_bar(notes, bars):
    """
    Return the bass note of each bar
    Parameters
    ----------
    notes
    bars

    Returns
    -------

    """
    drums_array = notes[:, 5] == 9
    notes = notes[~drums_array]
    start_times, end_times = notes[:, 0], notes[:, 0] + notes[:, 2]
    bass_notes = []
    for bar in bars:
        start, end = bar
        bar_notes = notes[((start_times >= start) & (start_times < end)) | ((end_times >= start) & (end_times < end))]
        if len(bar_notes) == 0:
            bass_notes.append(None)
        else:
            min_pitch = np.min(bar_notes[:, 1]) % 12
            bass_notes.append(min_pitch)
    return bass_notes


def dynamic_programming_for_chord_inference(lists):
    n = len(lists)
    m = max(len(lst) for lst in lists)
    dp = [[float('inf')] * m for _ in range(n)]
    indices = [[0] * m for _ in range(n)]

    for i, s in enumerate(lists[0]):
        dp[0][i] = 0  # No string changes needed at the start

    for i in range(1, n):
        for j in range(len(lists[i])):
            for k in range(len(lists[i - 1])):
                if lists[i][j] != lists[i - 1][k]:  # If the strings are different, we need a change
                    cost = 1
                else:
                    cost = 0

                if dp[i][j] > dp[i - 1][
                    k] + cost:  # If we found a better (smaller) cost, update the dp and indices array
                    dp[i][j] = dp[i - 1][k] + cost
                    indices[i][j] = k

    res = [0] * n
    min_cost, min_index = min(
        (cost, index) for index, cost in enumerate(dp[-1]))  # Find the smallest cost in the last row
    res[-1] = min_index  # The last index is the one with the smallest cost

    for i in range(n - 2, -1, -1):  # Backtrack to find the indices for the other lists
        res[i] = indices[i + 1][res[i + 1]]

    return res


def get_pitch_tonality_vector(tonality_root, tonality_mode):

    tonality = set([s % 12 for s in Tonality(tonality_root, tonality_mode).scale_pitches])
    pitch_vector = [1.0 * (i in tonality) for i in range(12)]
    return pitch_vector

def fast_chord_inference(notes, bars):
    """
    Return the chord progression of the song, one chord per bar
    It uses dynamic programming to find the roman numeral chord progression that minimize the number of tonality changes
    Parameters
    ----------
    notes
    bars

    Returns
    -------

    """
    chords = []
    bar_chroma_vectors = get_chroma_vectors(notes, bars)
    # Get the lowest note of each bar
    bass_notes = get_bass_note_bar(notes, bars)

    chord_roots, chord_types = max_correlation_index(bar_chroma_vectors, CHROMA_VECTORS_MATRIX)
    # Get maximum
    candidates = []
    chords = []
    for chord_root, chord_type, bar, chroma in zip(chord_roots, chord_types, bars, bar_chroma_vectors):

        from musiclang import Tonality, Chord
        # Get the list of roman numeral candidates
        pitches = tuple(sorted([(chord_root + pitch) % 12 for pitch in CHORD_TYPE_TO_PITCHES[CHORD_TYPES[chord_type]]]))
        subcandidates = PITCH_CLASSES_DICT[pitches]
        # Filter only maximum correlation candidates
        pitch_vectors = np.asarray([get_pitch_tonality_vector(tonality_root, tonality_mode)
                                    for (roman, tonality_root, tonality_mode, extension) in subcandidates])
        correlations = np.asarray([np.correlate(chroma, pitch_vector) for pitch_vector in pitch_vectors])
        # Get the list of maximum correlation candidates (there can be several)
        max_correlations = np.where(correlations == correlations.max())[0]
        subcandidates = [subcandidates[i] for i in max_correlations]
        candidates.append(subcandidates)


    # Choose the candidate chord path that minimize the number of tonality changes
    candidates_nb = [[(c[1] + 3 * (c[2] == 'm')) % 12 for c in candidate] for candidate in candidates]
    
    indexes = dynamic_programming_for_chord_inference(candidates_nb)
    chords_props = [candidates[i][indexes[i]] for i in range(len(indexes))]
    for (roman, tonality_root, tonality_mode, extension ), bar, bass_note in zip(chords_props, bars, bass_notes):
        chord = Chord(roman, tonality=Tonality(tonality_root, tonality_mode)).set_duration(bar[1] - bar[0])[extension]

        # Assign bass note if possible
        if bass_note is not None:
            chord_pitches = [pitch % 12 for pitch in chord.chord_pitches]
            nb_notes = len(chord_pitches)
            if bass_note in chord_pitches:
                bass_idx = chord_pitches.index(bass_note)
                EXTENSION_DICT = {(0, 3): '', (1, 3): '6', (2, 3): '64', (0, 4): '7', (1, 4): '65', (2, 4): '43', (3, 4): '2'}
                extension = EXTENSION_DICT[(bass_idx, nb_notes)]
                new_extension = extension + (chord.extension[1:] if chord.extension.startswith('7') else chord.extension)
                chord.extension = new_extension
        chords.append(chord)

    chords = Score(chords)
    return chords


"""The output tonal representations learned through multitask learning."""

import numpy as np

from .cache import (
    TransposeKey,
    TransposePcSet,
    TransposePitch,
)
from .feature_representation import (
    FeatureRepresentation,
    FeatureRepresentationTI,
    CHORD_QUALITIES,
    COMMON_ROMAN_NUMERALS,
    DEGREES,
    NOTEDURATIONS,
    KEYS,
    PCSETS,
    SPELLINGS,
)


class OutputRepresentation(FeatureRepresentation):
    """Output representations are all one-hot encoded (no many-hots).
    
    That makes them easier to template.

    Parameters
    ----------

    Returns
    -------

    """

    classList = []
    dfFeature = ""
    transpositionFn = None

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, dfFeature in enumerate(self.df[self.dfFeature]):
            transposed = self.transpositionFn(dfFeature, transposition)
            rnIndex = self.classList.index(transposed)
            array[frame] = rnIndex
        return array

    @classmethod
    def classesNumber(cls):
        """ """
        return len(cls.classList)

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        return [cls.classList[index] for index in array.reshape(-1)]

    @classmethod
    def decodeOneHot(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != len(cls.classList):
            raise IndexError("Strange array shape.")
        return [cls.classList[np.argmax(onehot)] for onehot in array]


class OutputRepresentationTI(FeatureRepresentationTI):
    """Output representations are all one-hot encoded (no many-hots).
    
    That makes them easier to template.

    Parameters
    ----------

    Returns
    -------

    """

    classList = []
    dfFeature = ""

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, dfFeature in enumerate(self.df[self.dfFeature]):
            rnIndex = self.classList.index(dfFeature)
            array[frame] = rnIndex
        return array

    @classmethod
    def classesNumber(cls):
        """ """
        return len(cls.classList)

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        return [cls.classList[index] for index in array.reshape(-1)]

    @classmethod
    def decodeOneHot(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != len(cls.classList):
            raise IndexError("Strange array shape.")
        return [cls.classList[np.argmax(onehot)] for onehot in array]


class Bass35(OutputRepresentation):
    """ """
    classList = SPELLINGS
    dfFeature = "a_bass"
    transpositionFn = staticmethod(TransposePitch)


class Tenor35(OutputRepresentation):
    """ """
    classList = SPELLINGS
    dfFeature = "a_tenor"
    transpositionFn = staticmethod(TransposePitch)


class Alto35(OutputRepresentation):
    """ """
    classList = SPELLINGS
    dfFeature = "a_alto"
    transpositionFn = staticmethod(TransposePitch)


class Soprano35(OutputRepresentation):
    """ """
    classList = SPELLINGS
    dfFeature = "a_soprano"
    transpositionFn = staticmethod(TransposePitch)


class Inversion4(OutputRepresentationTI):
    """ """
    classList = list(range(4))
    dfFeature = "a_inversion"

    def run(self):
        """ """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, inversion in enumerate(self.df[self.dfFeature]):
            if inversion > 3:
                # Any chord beyond sevenths is encoded as "root" position
                inversion = 0
            array[frame] = int(inversion)
        return array


class HarmonicRhythm7(OutputRepresentationTI):
    """ """
    classList = NOTEDURATIONS
    dfFeature = "a_harmonicRhythm"


class RomanNumeral31(OutputRepresentationTI):
    """ """
    classList = COMMON_ROMAN_NUMERALS
    dfFeature = "a_romanNumeral"


class PrimaryDegree22(OutputRepresentationTI):
    """ """
    classList = DEGREES
    dfFeature = "a_degree1"


class SecondaryDegree22(OutputRepresentationTI):
    """ """
    classList = DEGREES
    dfFeature = "a_degree2"


class LocalKey38(OutputRepresentation):
    """ """
    classList = KEYS
    dfFeature = "a_localKey"
    transpositionFn = staticmethod(TransposeKey)


class TonicizedKey38(OutputRepresentation):
    """ """
    classList = KEYS
    dfFeature = "a_tonicizedKey"
    transpositionFn = staticmethod(TransposeKey)


class ChordRoot35(OutputRepresentation):
    """ """
    classList = SPELLINGS
    dfFeature = "a_root"
    transpositionFn = staticmethod(TransposePitch)


class ChordQuality11(OutputRepresentationTI):
    """ """
    classList = CHORD_QUALITIES
    dfFeature = "a_quality"


class PitchClassSet121(OutputRepresentation):
    """ """
    classList = PCSETS
    dfFeature = "a_pcset"
    transpositionFn = staticmethod(TransposePcSet)


available_representations = {
    "Bass35": Bass35,
    "Tenor35": Tenor35,
    "Alto35": Alto35,
    "Soprano35": Soprano35,
    "ChordQuality11": ChordQuality11,
    "ChordRoot35": ChordRoot35,
    "HarmonicRhythm7": HarmonicRhythm7,
    "Inversion4": Inversion4,
    "LocalKey38": LocalKey38,
    "PitchClassSet121": PitchClassSet121,
    "PrimaryDegree22": PrimaryDegree22,
    "RomanNumeral31": RomanNumeral31,
    "SecondaryDegree22": SecondaryDegree22,
    "TonicizedKey38": TonicizedKey38,
}



from musiclang import Score
from musiclang.library import *
import os

import pytest

@pytest.mark.filterwarnings("ignore")
@pytest.mark.skip()
def test_read_file():

    filename = os.path.join(os.path.dirname(__file__), '../../examples/data/moonlight.mid')
    score = Score.from_midi(filename)

    assert score.chords[0].to_chord() == (I % II.b.m)



import json
from musiclang import Score
from .pattern_analyzer import PatternExtractor

class PicklePatternSampler:
    """
    Sample a chord pattern from a nested directory of musiclang text files
    """
    def __init__(self, directory, metadata_file=None,
                 bar_duration=4, min_nb_instruments=2,
                 **kwargs
                 ):
        self.directory = directory
        self.metadata_file = metadata_file
        self.bar_duration = bar_duration
        self.files = self.load_files()
        self.metadata = self.load_metadata()
        self.min_nb_instruments = min_nb_instruments
        self.kwargs = kwargs

    def load_metadata(self):
        if self.metadata_file is None:
            return {}
        else:
            with open(self.metadata_file, 'r') as f:
                metadata = json.load(f)
            return metadata

    def load_files(self):
        import glob
        files = glob.glob(self.directory + '/**/*.pkl', recursive=True)
        return files

    def sample(self):
        # Sample a file
        import random
        max_iter = 100
        idx = 0
        while idx < max_iter:
            file = random.choice(self.files)
            # Load the score from the file
            score = Score.from_pickle(file)

            # if score.config['time_signature'] != (4, 4):
            #     print('No good time signature', score.config['time_signature'])
            #     continue
            # Look for a chord with proper bar duration
            available_chords = [chord for chord in score.chords if (chord.duration == self.bar_duration)
                                and len(chord.instruments) >= self.min_nb_instruments]
            if len(available_chords) == 0:
                idx += 1
                continue

            chord_idx = random.choice(range(len(available_chords)))
            chord = available_chords[chord_idx]
            # Sort instruments from lower to higher pitch
            return PatternExtractor(**self.kwargs).extract(chord)

        raise Exception('No chord respecting criterias were found')



class TextPatternSampler(PicklePatternSampler):


    def load_files(self):
        import glob
        files = glob.glob(self.directory + '/**/*.txt', recursive=True)
        return files



from musiclang.library import *
from musiclang.transform.library import *
from musiclang.transform.composing.pattern import Nocturne, StandardPlucked, EpicMusicTernary
from musiclang.transform.composing import PartComposer
from musiclang.predict import compose_melody_and_harmony

def get_epic_orchestration(melody, harmony, tonality, bar_duration=6):
    composer = PartComposer(**{
        'orchestral_layers': [{
            "instruments": ['contrabasses', 'string_ensemble_1', 'french_horn'],
            "voicing": [b0.o(-2), b1.o(-1), b2.o(-1)],
            "orchestra": EpicMusicTernary(drums=0),
            "fixed_bass": True,
            "voice_leading": True
        },
            {

                "instruments": ['choir_aahs', 'choir_aahs', 'choir_aahs'],
                "voicing":  [b2, b0.o(1), b1.o(1)],
                "orchestra": StandardPlucked(bar_duration=bar_duration).to_json(),
                "fixed_bass": False,
                "voice_leading": True
            }
        ],
        'melodic_layers': [{
            "instrument": "string_ensemble_1",
            "melody": melody
        }],
        'global_layer': {
            "tonality": tonality,
            "harmony": harmony,
            "acc_amp": 'f',
            "patternator": {"restart_each_chord": False}
        }
    })
    return composer.to_json()



from musiclang import Chord, Tonality, Element, Silence, Note, Continuation
from musiclang.write.out.to_midi import *


def test_note_to_pitch_down():
    from musiclang import Chord, Tonality, Element

    note = Note("sd", 1, 0, 1)
    last_pitch = [12]
    chord = Element(3) % Tonality(2, mode="m", octave=-1)
    result = note_to_pitch(note, chord, 0, 12, last_pitch)
    res = [10, 12, 1, 66, 0, 0, 0, None, None]
    assert result == (res, res)


def test_note_to_pitch_up():
    note = Note("su", 1, 0, 1)
    last_pitch = [12]
    chord = Element(3) % Tonality(1, mode="m", octave=-1)
    result = note_to_pitch(note, chord, 0, 12, last_pitch)
    res = [13, 12, 1, 66, 0, 0, 0, None, None]
    assert result == (res, res)


def test_note_to_pitch_abs():
    from musiclang import Chord, Tonality, Element

    note = Note("s", 1, -1, 1)
    last_pitch = [12]
    chord = Element(3) % Tonality(2, mode="m", octave=-1)
    result = note_to_pitch(note, chord, 0, 12, last_pitch)
    res = [-15, 12, 1, 66, 0, 0, 0, None, None]
    assert result == (res, res)


def test_note_to_pitch_abs_octave_tonality():
    from musiclang import Chord, Tonality, Element

    note = Note("s", 1, -1, 1)
    last_pitch = [12]
    chord = Element(3).o(1) % Tonality(2, mode="m", octave=-1)
    result = note_to_pitch(note, chord, 0, 12, last_pitch)
    res = [-3, 12, 1, 66, 0, 0, 0, None, None]
    assert result == (res, res)


def test_note_to_pitch_silence():
    note = Silence(1)
    last_pitch = [12]
    chord = Element(3) % Tonality(1, mode="m", octave=-1)
    result = note_to_pitch(note, chord, 0, 12, last_pitch)
    res = [0, 12, 1, 66, 0, 1, 0, None, None]
    assert result == (res, last_pitch)


def test_note_to_pitch_continuation():
    note = Continuation(1)
    last_pitch = [12]
    chord = Element(3) % Tonality(1, mode="m", octave=-1)
    result = note_to_pitch(note, chord, 0, 12, last_pitch)
    res = [0, 12, 1, 66, 0, 0, 1, None, None]
    assert result == (res, last_pitch)


def test_to_midi():
    notes = [[0, 12, 1, 66, 0, 0, 0, None, None], [0, 12, 1, 66, 0, 0, 0, None, None]]
    res = to_midi(notes, output_file=None)
    assert res is not None


def test_score_to_midi():
    from musiclang.write.library import s0, s1, s2, I
    score = I(piano__0=s0 + s1 + s2)
    res = score_to_midi(score, None)
    assert res is not None


def test_get_track_list():
    from musiclang.write.library import s0, s1, s2, I
    chord1 = I(piano__0=s0 + s1 + s2, piano__1=s0 + s1 + s2, violin__0=s0 + s1 + s2)
    chord2 = I(piano__0=s0 + s1 + s2, oboe__0=s0 + s1 + s2, violin__0=s0 + s1 + s2)
    score = chord1 + chord2
    tracks = get_track_list(score)

    assert tracks == ['piano__0', 'piano__1', 'violin__0', 'oboe__0']


def test_tracks_to_instruments():
    res, names = tracks_to_instruments(['piano__0', 'violin__0'])
    assert res == {0: 0, 1: 40}
    assert names == ['piano', 'violin']


def test_create_melody_for_track():
    from musiclang.write.library import s0, s1, s2, I, V
    chord1 = I(piano__0=s0 + s1 + s2, piano__1=s0 + s1 + s2, violin__0=s0 + s1 + s2)
    chord2 = V(piano__0=s0 + s1 + s2, oboe__0=s0 + s1 + s2, violin__0=s0 + s1 + s2)
    score = chord1 + chord2

    res = create_melody_for_track(score, 0, 'piano__0')
    expected_res = [[0, 0, 1, 66, 0, 0, 0, None, None],
                    [2, 1, 1, 66, 0, 0, 0, None, None],
                    [4, 2, 1, 66, 0, 0, 0, None, None],
                    [7, 3, 1, 66, 0, 0, 0, None, None],
                    [9, 4, 1, 66, 0, 0, 0, None, None],
                    [11, 5, 1, 66, 0, 0, 0, None, None]]
    assert res == expected_res



"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

"""
This file groups a set of functions to parse files into MusicLang objects
"""

def parse_to_musiclang(input_file: str, **kwargs):
    """Parse an input file into a musiclang Score
    - Get chords with the AugmentedNet (https://github.com/napulen/AugmentedNet)
    - Get voice separation and parsing

    Parameters
    ----------
    input_file : str
        Input filepath

    Returns
    -------
    score: musiclang.Score
        Parsed score

    config: dict
        Dict of score configuration

    """
    extension = input_file.split('.')[-1]
    if extension.lower() in ['mid', 'midi']:
        return parse_midi_to_musiclang(input_file, **kwargs)
    elif extension.lower() in ['mxl', 'xml', 'musicxml', 'krn']:
        return parse_mxl_to_musiclang(input_file, **kwargs)
    else:
        raise Exception('Unknown extension {}'.format(extension))


def tokenize_midi_file(input_file, output_file, chord_range=None, quantization=16):
    import itertools
    def get_chord_range(tokens, tokenizer, start, end):
        bar_none = tokenizer['Bar_None']

        # Find first tempo token
        types = [event.type for event in tokens.events]
        # Find first Tempo event
        try:
            tempo = tokens[types.index('Tempo')]
        except Exception as e:
            tempo = None
        # Split tokens per bar_none token (like a str.split) (list to list of list)
        tokens = [list(g) for k, g in itertools.groupby(tokens, lambda x: x == bar_none) if not k]
        # Get the sublist between start and end
        tokens = tokens[start:end]
        # Add bar_none to each sublist
        tokens = [[bar_none] + t for t in tokens]
        # Flatten the list adding bar_none between each sublist
        tokens = list(itertools.chain.from_iterable(tokens))
        if tempo is not None and start > 0:
            tokens.insert(3, tempo)

        return tokens

    from miditok import REMI
    from miditok.classes import TokenizerConfig
    from miditoolkit import MidiFile

    config = TokenizerConfig(
        beat_res={(0, 8): quantization, (8, 16): quantization},
        use_tempos=True,
        use_time_signatures=True,
        one_token_stream_for_programs=True,
        use_programs=True)


    tokenizer = REMI(
        tokenizer_config=config,
    )
    #tokenizer.one_token_stream = True

    #[tokenizer.add_to_vocab(f'Position_{idx}') for idx in range(64, 256)]

    tokens = tokenizer.midi_to_tokens(MidiFile(input_file))

    if chord_range is not None:
        tokens = get_chord_range(tokens, tokenizer, *chord_range)


    midi = tokenizer.tokens_to_midi(tokens)

    # Reload for test
    midi.dump(output_file)

def parse_midi_to_musiclang(input_file: str, chord_range=None, tokenize_before=True, quantization=16, fast_chord_inference=True, **kwargs):
    """Parse a midi input file into a musiclang Score
    - Get chords with dynamic programming
    - Get voice separation and parsing

    Parameters
    ----------
    input_file : str
        Input midi filepath

    Returns
    -------
    score: musiclang.Score
        Parsed score

    config: dict
     Dict of score configuration

    """
    import tempfile
    import time
    import os
    import shutil
    # First tokenize the midi file
    with tempfile.TemporaryDirectory() as di:
        midi_file = os.path.join(di, 'data.mid')
        if tokenize_before:
            tokenize_midi_file(input_file, midi_file, chord_range=chord_range, quantization=quantization)
        elif chord_range is None:
            shutil.copy(input_file, midi_file)
        else:
            raise Exception('Chord range must be None if tokenize_before is False')

        if not fast_chord_inference:
            mxl_file = os.path.join(di, 'data.mxl')
            obj = _m21Parse(midi_file)
            obj.write('mxl', fp=os.path.join(mxl_file))

        result = parse_directory_to_musiclang(di, fast_chord_inference=fast_chord_inference, **kwargs)
    return result


def parse_mxl_to_musiclang(input_file: str, **kwargs):
    """Parse a music xml input file into a musiclang Score
    - Get chords with the AugmentedNet (https://github.com/napulen/AugmentedNet)
    - Separate into monophonic voice with the proper instrument

    Parameters
    ----------
    input_file: str :
            Filepath of the input

    Returns
    -------
    score : musiclang.Score

    config: dict
        Dict of score configuration

    """

    import tempfile
    import os
    import music21
    import shutil
    with tempfile.TemporaryDirectory() as di:
        midi_file = os.path.join(di, 'data.mid')
        mxl_file = os.path.join(di, 'data.mxl')
        obj = _m21Parse(input_file)
        obj.write('midi', fp=os.path.join(midi_file))
        shutil.copy(input_file, mxl_file)
        result = parse_directory_to_musiclang(di, **kwargs)
    return result

def parse_directory_to_musiclang(directory: str, fast_chord_inference=True, **kwargs):
    """Parse a directory containing a 'data.mid' and 'data_annotated.rntxt' file (midi file and chord annotation file)

    Parameters
    ----------
    directory : str
        Directory with a "data.mid" file and a "data_annotated.rntxt" annotation file

    Returns
    -------
    score: Score
        MusicLang score parsed

    config: dict
        Dict of score configuration

    """
    import os
    midi_file = os.path.join(directory, 'data.mid')
    mxl_file = os.path.join(directory, 'data.mxl')
    if not fast_chord_inference:
        from .augmented_net import infer_chords
        pprint('1/4 : Analyze the score (This may takes a while)')
        annotation_file = os.path.join(directory, 'data_annotated.rntxt')
        infer_chords(mxl_file)
        score, config = parse_midi_to_musiclang_with_annotation(midi_file, annotation_file)

    else:
        score, config = parse_midi_to_musiclang_without_annotation(midi_file)

    score = score.to_drum()
    return score, config



def parse_midi_to_musiclang_without_annotation(midi_file: str):
    """

    Parameters
    ----------
    midi_file: str :
        Filepath to the midi file to parse

    annotation_file: str :
        Filepath to the anotation file to parse


    Returns
    -------
    score: Score
        MusicLang score parsed

    config: dict
        Config of the score

    """
    config = {}
    score, tempo = parse_musiclang_sequence_and_chords(midi_file)

    config.update({'tempo': tempo, 'time_signature': score.config['time_signature']})
    return score, config

def parse_midi_to_musiclang_with_annotation(midi_file: str, annotation_file: str):
    """

    Parameters
    ----------
    midi_file: str :
        Filepath to the midi file to parse

    annotation_file: str :
        Filepath to the anotation file to parse


    Returns
    -------
    score: Score
        MusicLang score parsed

    config: dict
        Config of the score

    """


    chords = get_chords_from_analysis(annotation_file)
    config = chords.config
    score, tempo = parse_musiclang_sequence(midi_file, chords)
    config.update({'tempo': tempo})
    return score, config


# Helpers for parser


def parse_roman_numeral(element):
    print(element.primaryFigure + element.secondaryRomanNumeral.primaryFigure)

def parse_tonality(element):
    """Parse tonality as written in musicLang from Music21 object

    Parameters
    ----------
    element :
        return: key_tonic : Degree of the key (in 0-12)

    Returns
    -------
    type
        key_tonic : Degree of the key (in 0-12)

    """

    DICT_MODE = {'major': 'M', 'minor': 'm'}
    if element.secondaryRomanNumeralKey is not None:
        key_tonic = element.secondaryRomanNumeralKey.tonic.pitchClass
        key_mode = DICT_MODE[element.secondaryRomanNumeralKey.mode]
    else:
        key_tonic = element.key.tonic.pitchClass
        key_mode = DICT_MODE[element.key.mode]

    figure = element.primaryFigure

    if 'N' in figure:
        key_tonic += 1
        key_mode = 'M'
    elif 'Ger' in figure:
        key_tonic += 1  # For musiclang it will be a V % II.b.M
        key_mode = 'M'
    elif 'It' in figure:
        key_tonic += 1  # For musiclang it will be a I % II.b.M
        key_mode = 'M'
    elif 'Fr' in figure:
        key_tonic += 3
        key_mode = 'mm'
    # elif 'bIII' in figure:
    #     key_tonic += 3
    #     key_mode = 'M'
    # elif 'biii' in figure:
    #     key_tonic += 3
    #     key_mode = 'm'
    # elif 'bIV' in figure:
    #     key_tonic += 4
    #     key_mode = 'M'
    # elif 'biv' in figure:
    #     key_tonic += 4
    #     key_mode = 'm'
    # elif 'bVII' in figure:
    #     key_tonic += 10
    #     key_mode = 'M'
    # elif 'bvii' in figure:
    #     key_tonic += 10
    #     key_mode = 'm'
    # elif 'bVI' in figure:
    #     key_tonic += 8
    #     key_mode = 'M'
    # elif 'bvi' in figure:
    #     key_tonic += 8
    #     key_mode = 'm'
    # elif 'bII' in figure:
    #     key_tonic += 1
    #     key_mode = 'M'
    # elif 'bii' in figure:
    #     key_tonic += 1
    #     key_mode = 'm'
    # elif 'bV' in figure:
    #     key_tonic += 6
    #     key_mode = 'M'
    # elif 'bv' in figure:
    #     key_tonic += 6
    #     key_mode = 'm'

    return key_tonic, key_mode


def get_degree(element):
    """Get the scale degree of an element (from Music21)

    Parameters
    ----------
    element :
        Music21 element

    Returns
    -------
    degree : int

    """
    degree = element.scaleDegree - 1
    if 'Ger' in element.primaryFigure:
        degree = 4
    elif 'It' in element.primaryFigure:
        degree = 4
    elif 'N' in element.primaryFigure:
        degree = 0
    elif 'Fr' in element.primaryFigure:
        degree = 3
    # elif 'bIII' in element.primaryFigure:
    #     degree = 0
    # elif 'biii' in element.primaryFigure:
    #     degree = 0
    # elif 'bIV' in element.primaryFigure:
    #     degree = 0
    # elif 'biv' in element.primaryFigure:
    #     degree = 0
    # elif 'bVII' in element.primaryFigure:
    #     degree = 0
    # elif 'bvii' in element.primaryFigure:
    #     degree = 0
    # elif 'bVI' in element.primaryFigure:
    #     degree = 0
    # elif 'bvi' in element.primaryFigure:
    #     degree = 0
    # elif 'bII' in element.primaryFigure:
    #     degree = 0
    # elif 'bii' in element.primaryFigure:
    #     degree = 0
    # elif 'bV' in element.primaryFigure:
    #     degree = 0
    # elif 'bv' in element.primaryFigure:
    #     degree = 0

    return degree



def get_duration(roman):
    """Get the duration of a chord

    Parameters
    ----------
    roman :


    Returns
    -------
    type
        result, fractions.Fraction with denominator limited to 8

    """
    from fractions import Fraction as frac
    return frac(roman.duration.quarterLength).limit_denominator(8)



def pprint(text):
    import os
    verbose = os.getenv('VERBOSE', False)
    if verbose:
        print(text)

def parse_musiclang_sequence_and_chords(midi_file):
    """Parse a midi file into MusicLang and chords

    Parameters
    ----------
    midi_file : str
        Path to the midi file
    """
    from musiclang import Score
    from .midi_parser import parse_midi
    from .item import convert_to_items
    from .to_musiclang import infer_score_with_chords_durations
    from .chord_inference import fast_chord_inference

    pprint('1/4 : Performing voice separation (This may takes a while)')
    notes, instruments, tempo, time_signature, bars = parse_midi(midi_file)

    pprint('2/4 : Now infering chords with fast inference')
    chords = fast_chord_inference(notes, bars)
    sequence = convert_to_items(notes)
    pprint('3/4 : Create the score')
    import time
    start = time.time()
    score = infer_score_with_chords_durations(sequence, chords, instruments)
    pprint('Finished creating score in {} seconds'.format(time.time() - start))
    pprint('Finished creating score')
    pprint('4/4 Normalize the score...')
    score = score.normalize()
    score.config['time_signature'] = time_signature

    return score, tempo


def parse_musiclang_sequence(midi_file, chords):
    """Parse a midi file into MusicLang and chords with chords duration

    Parameters
    ----------
    midi_file :

    chords :


    Returns
    -------

    """
    from musiclang import Score
    from .midi_parser import parse_midi
    from .item import convert_to_items
    from .to_musiclang import infer_score_with_chords_durations


    pprint('2/4 : Performing voice separation (This may takes a while)')
    notes, instruments, tempo, time_signature, bars = parse_midi(midi_file)
    sequence = convert_to_items(notes)
    pprint('3/4 : Create the score')
    score = infer_score_with_chords_durations(sequence, chords, instruments)
    pprint('Finished creating score')
    pprint('4/4 Normalize the score...')
    score = score.normalize()

    return score, tempo


def get_chords_from_mxl(input_file):
    """Given a mxl file, return a MusicLang score parsing only the chord progression

    Parameters
    ----------
    input_file : str
        MusicXML file to parse

    Returns
    -------
    score: musiclang.Score

    """
    import shutil
    import tempfile
    import os
    from .augmented_net import infer_chords
    with tempfile.TemporaryDirectory() as di:
        mxl_file = os.path.join(di, 'data.mxl')
        annotated_file = os.path.join(di, 'data_annotated.rntxt')
        shutil.copy(input_file, mxl_file)
        infer_chords(mxl_file)
        return get_chords_from_analysis(annotated_file)


def get_chords_from_analysis(analysis):
    """Given an analysis file (.rntxt), return a MusicLang score parsing only the chord progression

    Parameters
    ----------
    analysis : str,
        Filepath of analysis file (.rntxt) format


    Returns
    -------
    score: musiclang.Score

    """
    from .score_formatter import ScoreFormatter
    with open(analysis, 'r') as f:
        chords = ScoreFormatter(f.read()).parse()
    return chords



def chords_to_musiclang(chords):
    """

    Parameters
    ----------
    chords :


    Returns
    -------

    """
    from musiclang import Tonality, Chord, Note
    new_score = None
    for notes, duration, degree, figure, key in chords:
        key_tonic, key_mode = key
        tonality = Tonality(key_tonic, mode=key_mode)
        chord = Chord(degree, tonality=tonality, extension=figure)
        scale_notes = []
        for pitch_class, octave in notes:
            # Parse the note in musiclang
            scale_notes.append(chord.parse(pitch_class).o(octave + 1).augment(duration))

        chord_score = {f'piano__{idx}': sn for idx, sn in enumerate(scale_notes)}
        new_score += chord(**chord_score)

    return new_score


def music21_roman_analysis_to_chords(score):
    """

    Parameters
    ----------
    score :


    Returns
    -------

    """
    time_sig = None

    chords = []
    # Get time sigs
    romans = score.recurse().getElementsByClass('RomanNumeral')
    for roman in romans:
        key_tonic, key_mode = parse_tonality(roman)
        notes = [(n.pitch.pitchClass, n.pitch.octave - 5) for n in roman.notes]
        duration = get_duration(roman)
        degree = get_degree(roman)
        figure = roman.primaryFigure
        figure = figure.replace('bI', '').replace('bV', '')
        figure = figure.replace('#I', '').replace('#V', '')
        figure = figure.replace('I', '').replace('V', '')
        figure = figure.replace('bi', '').replace('bv', '')
        figure = figure.replace('#i', '').replace('#v', '')
        figure = figure.replace('i', '').replace('v', '')

        #primary_figure = roman.primaryFigure

        replacer = (
            ('M7', '[M7]'),
            ('maj7', '[M7]'),
            ('m7', '[m7]'),
            ('min7', '[m7]'),
            ('+', '(+)'),
            ('b5', '(b5)'),
            ('b7', '[m7]'),
            ('b9', '[m9]'),
            ('ar', ''),
            ('add2', '[add2]'),
            ('add4', '[add4]'),
            ('add6', '[add6]'),
            ('add9', '[add9]'),
            ('add11', '[add11]'),
            ('add13', '[add13]'),
            ('sus2', '(sus2)'),
            ('sus4', '(sus4)'),
            ('[no5]', '{-5}'),
            ('[no3]', '{-3}'),
            ('[no1]', '{-1}'),
            ('[no7]', '{-7}'),
        )

        figure = figure.replace('[', '').replace('(', '').replace(']', '').replace(')', '')
        for key, val in replacer:
            figure = figure.replace(key, val)

        chords.append([notes, duration, degree, figure, (key_tonic, key_mode)])

    return chords


def old_annotation_to_musiclang(file):
    import music21
    analysis = music21.converter.parse(file, format="romanText")
    chords = music21_roman_analysis_to_chords(analysis)
    score = chords_to_musiclang(chords)
    return score


def _m21Parse(f, fmt=None, remove_perc=True):
    """

    Parameters
    ----------
    f :

    fmt :
         (Default value = None)

    Returns
    -------

    """
    import music21
    s = music21.converter.parse(f, format=fmt, forceSource=True)
    if remove_perc:

        for el in s.recurse():
            if set(el.classes).intersection(['PercussionChord', 'Unpitched']):  # or 'Piano'
                # el.activeSite.replace(el, instrument.Violin())
                el.activeSite.remove(el)

        # perc = [
        #     p
        #     for p in s.parts
        #     if list(p.recurse().getElementsByClass(["PercussionClef", "PercussionChord"]))
        # ]
        # s.remove(perc, recurse=True)
    return s



from musiclang.transform.composing import create_counterpoint, create_counterpoint_on_score
from musiclang.library import *

def test_create_counterpoint():
    melody1 = s0 + s1 + s2
    melody2 = s0 + s1 + s2  # Parallel unissons, should be resolved

    melody2_corrected = create_counterpoint([melody1], [melody2])

    assert melody2_corrected[0] != s0 + s1 + s2


from typing import Dict
from .metric import Metric

class ScoreRhythm:
    """
    Helper class to apply rhythms
    """

    def __init__(self, rhythm_dict):
        self.rhythm_dict: Dict[Metric] = rhythm_dict

    def signature(self, part):
        return self.rhythm_dict[part].signature

    def tatum(self, part):
        return self.rhythm_dict[part].tatum

    def array(self, part):
        return self.rhythm_dict[part].array

    def nb_bars(self, part):
        return self.rhythm_dict[part].nb_bars

    def duration(self, part):
        return self.rhythm_dict[part].duration

    def __call__(self, score, **kwargs):
        """
        Apply the rythm dict to the score, keeping durations

        Parameters
        ----------
        score
        kwargs

        Returns
        -------
        """

        new_score = None
        time = 0
        for chord in score.chords:
            duration = chord.duration
            new_part = {}
            for part in chord.parts:
                if part in self.rhythm_dict:
                    bar_time_start = time
                    bar_time_end = (time + chord.duration)
                    if bar_time_end == 0:
                        bar_time_end = self.duration(part)
                    assert bar_time_end - bar_time_start > 0, str((bar_time_end, bar_time_start, self.duration(part)))
                    try:
                        temps_melody = self.rhythm_dict[part].apply_to_melody(chord.score[part], start=bar_time_start, end=bar_time_end)
                        melody = temps_melody
                    except IndexError:
                        melody = chord.score[part].augment(chord.duration/chord.score[part].duration).to_melody()

                    assert melody.duration == chord.duration


                else:
                    melody = chord.score[part]
                new_part[part] = melody

            new_score += chord(**new_part)
            time += duration

        return new_score




"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

class NoteProperties:
    """ """

    def __init__(self, note):
        self.note = note


    @property
    def notes(self):
        """ """
        from ..melody import Melody
        return Melody([self.note], tags=self.note.tags).notes

    @property
    def scale_pitch(self):
        """ """
        if self.note.type == "s":
            return self.note.val + 7 * self.note.octave
        elif self.note.type == "h":
            return (7 * self.note.val//12) + 7 * self.note.octave
        else:
            raise Exception(f'Not well defined pitch for type {self.note.type}')

    @property
    def delta_value(self):
        """ """
        if self.note.type[1] == "u":
            return self.note.val
        elif self.note.type[1] == "d":
            return -self.note.val
        else:
            raise Exception(f'Not well defined delta value for type {self.note.type}')

    @property
    def amp_normalized(self):
        return self.note.amp / 120


    @property
    def n(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.0
        return note

    @property
    def ppp(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.16
        return note

    @property
    def pp(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.26
        return note

    @property
    def p(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.36
        return note

    @property
    def mp(self):
        note = self.note.copy()
        note.amp = 120 * 0.5
        return note

    @property
    def mf(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.65
        return note

    @property
    def f(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.8
        return note

    @property
    def ff(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.9
        return note

    @property
    def fff(self):
        """ """
        note = self.note.copy()
        note.amp = 120 * 0.95
        return note

    @property
    def amp_figure(self):
        """ """
        n = self.note.amp_normalized
        if n is None or n <= 0:
            return 'n'
        elif n <= 0.16:
            return 'ppp'
        elif n <= 0.26:
            return 'pp'
        elif n <= 0.36:
            return 'p'
        elif n <= 0.5:
            return 'mp'
        elif n <= 0.65:
            return 'mf'
        elif n <= 0.8:
            return 'f'
        elif n <= 0.9:
            return 'ff'
        else:
            return 'fff'

    @property
    def is_up(self):
        """ """
        return "u" in self.note.type[1:]

    @property
    def is_down(self):
        """ """
        return "d" in self.note.type[1:]

    @property
    def had_absolute_note(self):
        """ """
        return self.note.starts_with_absolute_note

    @property
    def is_relative(self):
        """ """
        return (self.note.type != "d") and ("u" in self.note.type or "d" in self.note.type)

    @property
    def starts_with_relative(self):
        """ """
        return self.note.is_relative

    @property
    def starts_with_absolute_or_silence(self):
        """ """
        return self.note.type == "r" or self.note.starts_with_absolute_note

    @property
    def is_silence(self):
        """ """
        return self.note.type == "r"

    @property
    def is_drum_note(self):
        """ """
        return "d" in self.note.type

    @property
    def is_pattern_note(self):
        """ """
        return self.note.type == "a"

    @property
    def is_chromatic_note(self):
        """ """
        return self.note.type in ["h", "hu", "hd"]

    @property
    def is_scale_note(self):
        """ """
        return self.note.type in ["s", "su", "sd"]

    @property
    def is_chord_note(self):
        """ """
        return self.note.type in ['c', 'cu', 'cd']

    @property
    def is_absolute_note(self):
        return self.note.type in ['a', 'au', 'ad']

    @property
    def is_bass_note(self):
        """ """
        return self.note.type in ['b', 'bu', 'bd']

    @property
    def is_continuation(self):
        """ """
        return self.note.type == "l"

    @property
    def is_note(self):
        """ """
        return self.note.type not in ["r", "l", "d", "x"]

    @property
    def starts_with_note(self):
        """ """
        return self.note.type not in ["r", "l", "d", "x"]

    @property
    def starts_with_absolute_note(self):
        """ """
        return self.note.starts_with_note and not self.note.is_relative

    @property
    def M(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "M"
        return new_note

    @property
    def m(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "m"
        return new_note

    @property
    def mm(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "mm"
        return new_note

    @property
    def dorian(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "dorian"
        return new_note

    @property
    def phrygian(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "phrygian"
        return new_note

    @property
    def lydian(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "lydian"
        return new_note

    @property
    def mixolydian(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "mixolydian"
        return new_note

    @property
    def aeolian(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "aeolian"
        return new_note

    @property
    def locrian(self):
        """ """
        new_note = self.note.copy()
        new_note.mode = "locrian"
        return new_note

    @property
    def dim(self):
        """ """
        new_note = self.note.copy()
        new_note.accident = "dim"
        return new_note

    @property
    def min(self):
        """ """
        new_note = self.note.copy()
        new_note.accident = "min"
        return new_note

    @property
    def natural(self):
        """ """
        new_note = self.note.copy()
        new_note.accident = "natural"
        return new_note

    @property
    def maj(self):
        """ """
        new_note = self.note.copy()
        new_note.accident = "maj"
        return new_note


    @property
    def aug(self):
        """ """
        new_note = self.note.copy()
        new_note.accident = "aug"
        return new_note


from music21 import note, stream, metadata, meter, tie, dynamics
import music21
from musiclang import Chord


MUSESCORE_REPLACE_DICT = {
    'French horn': 'Horns in F'
}


SCALES_MAJOR = {
    0: ['C', 'D', 'E', 'F', 'G', 'A', 'B'],
    1: ['Db', 'Eb', 'F', 'Gb', 'Ab', 'Bb', 'C'],
    2: ['D', 'E', 'F#', 'G', 'A', 'B', 'C#'],
    3: ['Eb', 'F', 'G', 'Ab', 'Bb', 'C', 'D'],
    4: ['E', 'F#', 'G#', 'A', 'B', 'C#', 'D#'],
    5: ['F', 'G', 'A', 'Bb', 'C', 'D', 'E'],
    6: ['Gb', 'Ab', 'Bb', 'Cb', 'Db', 'Eb', 'F'],
    7: ['G', 'A', 'B', 'C', 'D', 'E', 'F#'],
    8: ['Ab', 'Bb', 'C', 'Db', 'Eb', 'F', 'G'],
    9: ['A', 'B', 'C#', 'D', 'E', 'F#', 'G#'],
    10: ['Bb', 'C', 'D', 'Eb', 'F', 'G', 'A'],
    11: ['B', 'C#', 'D#', 'E', 'F#', 'G#', 'A#']
}


SCALES_MINOR = {
    0: ['C', 'D', 'Eb', 'F', 'G', 'Ab', 'B'],
    1: ['C#', 'D#', 'E', 'F#', 'G#', 'A', 'B#'],
    2: ['D', 'E', 'F', 'G', 'A', 'Bb', 'C#'],
    3: ['Eb', 'F', 'Gb', 'Ab', 'Bb', 'Cb', 'D'],
    4: ['E', 'F#', 'G', 'A', 'B', 'C', 'D#'],
    5: ['F', 'G', 'Ab', 'Bb', 'C', 'Db', 'E'],
    6: ['F#', 'G#', 'A', 'B', 'C#', 'D', 'E#'],
    7: ['G', 'A', 'Bb', 'C', 'D', 'Eb', 'F#'],
    8: ['G#', 'A#', 'B', 'C#', 'D#', 'E', 'F##'],
    9: ['A', 'B', 'C', 'D', 'E', 'F', 'G#'],
    10: ['Bb', 'C', 'Db', 'Eb', 'F', 'Gb', 'A'],
    11: ['B', 'C#', 'D', 'E', 'F#', 'G', 'A#']
}

SCALES_MELODIC_MINOR = {
    0: ['C', 'D', 'Eb', 'F', 'G', 'A', 'B'],
    1: ['C#', 'D#', 'E', 'F#', 'G#', 'A#', 'B#'],
    2: ['D', 'E', 'F', 'G', 'A', 'B', 'C#'],
    3: ['Eb', 'F', 'Gb', 'Ab', 'Bb', 'C', 'D'],
    4: ['E', 'F#', 'G', 'A', 'B', 'C#', 'D#'],
    5: ['F', 'G', 'Ab', 'Bb', 'C', 'D', 'E'],
    6: ['F#', 'G#', 'A', 'B', 'C#', 'D#', 'E#'],
    7: ['G', 'A', 'Bb', 'C', 'D', 'E', 'F#'],
    8: ['G#', 'A#', 'B', 'C#', 'D#', 'E#', 'F##'],
    9: ['A', 'B', 'C', 'D', 'E', 'F#', 'G#'],
    10: ['Bb', 'C', 'Db', 'Eb', 'F', 'G', 'A'],
    11: ['B', 'C#', 'D', 'E', 'F#', 'G#', 'A#']
}

SCALES = {
    'm': SCALES_MINOR,
    'mm': SCALES_MELODIC_MINOR,
    'M': SCALES_MAJOR
}


def get_note_spelling(note, chord, last_pitch=None):
    from .to_midi import note_to_pitch_result
    degree = chord.degree
    tonality_degree = chord.tonality.degree
    mode = chord.tonality.mode
    pitch = note_to_pitch_result(note, chord, last_pitch=last_pitch)
    tonality_scale_pitches = [p % 12 for p in chord.tonality.scale_pitches]
    if (pitch % 12) in tonality_scale_pitches:
        idx = tonality_scale_pitches.index(pitch % 12)
        note_spelling = SCALES[mode][tonality_degree][idx]
        # Correct octave
        octave = (pitch + 48) // 12
        if note_spelling in ['B#']:
            octave -= 1
        elif note_spelling in ['Cb']:
            octave += 1
        note_spelling = note_spelling + str(octave)
        new_note = music21.note.Note(note_spelling)
        new_note.duration = music21.duration.Duration(note.duration)
    else:
        new_note = music21.note.Note(pitch % 12)
        new_note.octave = (pitch + 48) // 12
        new_note.duration = music21.duration.Duration(note.duration)

    return new_note, pitch


def tonality_to_music21_key(tonality):
    tonality_mode = tonality.mode
    tonality_degree = tonality.degree
    tonalities_major = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'F#', 'G', 'Ab', 'A', 'Bb', 'B']
    tonalities_minor = ['c', 'c#', 'd', 'd#', 'e', 'f', 'f#', 'g', 'g#', 'a', 'bb', 'b']
    tonalities = tonalities_major if tonality_mode in ['M', 'lydian', 'mixolydian'] else tonalities_minor
    key = music21.key.Key(tonalities[tonality_degree])
    return key


def find_tonality(score):
    from musiclang.transform.features import ExtractMainTonality
    tonality = ExtractMainTonality()(score)
    return tonality_to_music21_key(tonality)


def get_musescore_name(ins):
    """
    Get the name of the part as printed on the score
    Parameters
    ----------
    ins: str
        Raw instrument name

    Returns
    -------
    result: str
        Name
    """
    res = ins.capitalize().replace('_', ' ')

    return MUSESCORE_REPLACE_DICT.get(res, res)


def find_instruments(score):
    """
    Given a score return the list of instruments as printed on the score
    Parameters
    ----------
    score

    Returns
    -------
    parts_dict_names: dict
        Dict of music21 part for each instrument, keyed by part name in musiclang
    parts_dic: dict
        Dict of music21 part for each instrument keyed by instrument name

    """
    from music21 import instrument, stream
    from musiclang.write.out.constants import INSTRUMENTS_DICT
    from musiclang.write.constants import ALL_INST
    instruments = [(ins.split('__')[0], ins) for ins in score.instruments]

    parts_dict = {}
    parts_dict_names = {}
    # Regroup same instruments on same part
    for ins, part_name in instruments:
        if ins not in parts_dict:
            parts_dict[ins] = stream.Part()
            instr = instrument.Instrument(get_musescore_name(ins))
            instr.midiProgram = INSTRUMENTS_DICT.get(ins, 0)
            part = parts_dict[ins]
            part.append(instr)

        part = parts_dict[ins]
        parts_dict_names[part_name] = part

    return parts_dict_names, parts_dict

def score_instrument_to_notes(score, part_name, voice, voice_idx, no_repeat=False):
    """
    Given a score and a part name, returns the music21 voice with all the notes

    Parameters
    ----------
    score: Score
    part_name: str
    voice: music21.Voice
    voice_idx: int

    Returns
    -------
    voice: music21.Voice

    """
    last_spelling = None
    last_pitch = None
    last_is_silence = True
    curr_dynamic = 'mf'
    for chord in score.chords:
        voice, last_spelling, curr_dynamic, last_pitch, last_is_silence = chord_instrument_to_notes(chord, voice, part_name,
                                                                       voice_idx,
                                                                       last_spelling=last_spelling,
                                                                       curr_dynamic=curr_dynamic,
                                                                       no_repeat=no_repeat,
                                                                       last_pitch=last_pitch,
                                                                       last_is_silence=last_is_silence
                                                                       )
    return voice


def chord_to_musescore_lyric(chord: Chord):
    """
    Create the chord scale symbol on the score

    Parameters
    ----------
    chord

    Returns
    -------
    result: str
        Chord scale symbol
    """
    return "{} /{}".format(chord.element_to_str(), chord.tonality_to_str()).replace('%', '')


def chord_instrument_to_notes(chord, voice, part_name, ins_idx, last_spelling=None, curr_dynamic='mf', no_repeat=False,
                              last_pitch=None, last_is_silence=True):
    """
    Given a chord, get spellings

    Returns
    -------
    """

    # Enharmonic
    old_last_is_silence = last_is_silence
    last_is_silence = False
    FIGURES = ['n', 'ppp', 'pp', 'p', 'mp', 'mf', 'f', 'ff', 'fff']
    if part_name in chord.score.keys():
        part = chord.score[part_name]
        for idx_note, n in enumerate(part.notes):
            if n.is_note:
                old_last_pitch = last_pitch
                new_note, last_pitch = get_note_spelling(n, chord, last_pitch=last_pitch)
                last_spelling = new_note.nameWithOctave
                if idx_note == 0 and ins_idx == 0:
                    new_note.addLyric(chord_to_musescore_lyric(chord))
                if n.amp_figure != curr_dynamic:
                    dyn = dynamics.Dynamic(n.amp_figure)
                    voice.append(dyn)
                    curr_dynamic = n.amp_figure
                if (last_pitch != old_last_pitch) or (not no_repeat):
                    voice.append(new_note)
                else:
                    try:
                        if last_spelling is not None and not last_is_silence:
                            new_note = note.Note(last_spelling)
                            new_note.duration = music21.duration.Duration(n.duration)
                            voice[-1].tie = tie.Tie('start')
                            new_note.tie = tie.Tie('stop')
                            voice.append(new_note)
                        else:
                            voice.append(note.Rest(n.duration))
                            last_is_silence = True
                    except:
                        voice.append(note.Rest(n.duration))
                        last_is_silence = True
            elif n.is_silence:
                voice.append(note.Rest(n.duration))
                last_is_silence = True
            elif n.is_continuation:
                if old_last_is_silence:
                    last_is_silence = True
                try:
                    if last_spelling is not None and not last_is_silence:
                        new_note = note.Note(last_spelling)
                        new_note.duration = music21.duration.Duration(n.duration)
                        voice[-1].tie = tie.Tie('start')
                        new_note.tie = tie.Tie('stop')
                        voice.append(new_note)
                    else:
                        voice.append(note.Rest(n.duration))
                        last_is_silence = True
                except:
                    voice.append(note.Rest(n.duration))
                    last_is_silence = True

    else:
        voice.append(note.Rest(chord.duration))

    return voice, last_spelling, curr_dynamic, last_pitch, last_is_silence


def score_to_music_21(score, signature=(4, 4), tempo=50, tonality=None, title='MusicLang score', composer='MusicLang', no_repeat=False, **kwargs):
    """
    Transform a musiclang score into a Music21 score
    Parameters
    ----------
    score: Score
            MusicLang score to transform
    signature: tuple (nom, den)
            Time signature of the piece
    tempo: int
            Tempo of the piece
    tonality: Tonality
            Tonality of the piece if applicable
    title: str
        Title of the piece

    Returns
    -------
    m21_score: music21.Score
        Score transformed into a music21 object

    """
    # Find tonality
    if tonality is None:
        tonality = find_tonality(score)
    else:
        tonality = tonality_to_music21_key(tonality)

    # Find instruments
    parts, instruments = find_instruments(score)

    idx = 0
    # Add time signatures and tonalities ...
    for part_name, part in instruments.items():
        if idx == 0:
            part.append(music21.tempo.MetronomeMark('', tempo, note.Note(type='quarter')))
        part.append(tonality)
        part.append(meter.TimeSignature(f'{signature[0]}/{signature[1]}'))
        idx += 1

    # Create voices
    idx = 0
    for part_name, part in parts.items():
        # Add notes for each chord if first add lyrics ...
        voice = stream.Voice(number=idx + 1)
        voice = score_instrument_to_notes(score, part_name, voice, idx, no_repeat=no_repeat)

        part.insert(0, voice)
        idx += 1

    m21_score = stream.Score()
    meta = metadata.Metadata(title=title, composer=composer)
    m21_score.insert(0, meta)
    for part_name, part in instruments.items():
        m21_score.insert(0, part)
    # Iterate over chord for each instrument
    return m21_score


def score_to_mxl(score, filepath, signature=(4, 4), tempo=50, tonality=None, no_repeat=False, **kwargs):
    """
    Transform a musiclang score into a musicxml file, readable by all the main notation software (musescore, finale ...)

    Parameters
    ----------
    score: Score
            MusicLang score to transform
    filepath: str
            Filepath of the musicxml file
    signature: tuple (nom, den)
            Time signature of the piece
    tempo: int
            Tempo of the piece
    tonality: Tonality
            Tonality of the piece if applicable
    title: str
        Title of the piece

    """
    # Find tonality
    m21_score = score_to_music_21(score, signature=signature, tempo=tempo, tonality=tonality, no_repeat=no_repeat, **kwargs)
    m21_score.write('mxl', filepath)




from musiclang.library import *

def test_chord_note_basics():

    chord = (I % I.M)
    assert chord.to_pitch(c0) == 0
    chord = (I % I.M)['6']
    assert chord.to_pitch(c0) == 0
    assert chord.to_pitch(c1) == 4
    assert chord.to_pitch(c2) == 7
    assert chord.to_pitch(c3) == 12
    assert chord.to_pitch(c0.o(1)) == 12
    assert chord.to_pitch(c0.o(-1)) == -12



def test_chord_note_complicated_chord():

    chord = (II % II.M)['2[add2]']
    assert chord.to_pitch(c0) == 4
    assert chord.to_pitch(c1) == 6
    assert chord.to_pitch(c2) == 7
    assert chord.to_pitch(c3) == 11


def test_some_chords_works():
    D_major_sus2_second_inversion = (I % II.M)['64(sus2)']  # I in key of D major
    assert D_major_sus2_second_inversion.extension_notes == [s4, s0.o(1), s1.o(1)]
    augmented = (I % II.M)['(+)']
    assert augmented.extension_notes == [s0, s2, h8]
    add_sixth = (I % II.M)['[add6]']
    assert add_sixth.extension_notes == [s0, s2, s4, s5]


import pandas as pd
import numpy as np

COLUMNS_CHORDS = ['chord_degree', 'tonality_degree', 'tonality_mode']
COLUMNS_MELODY = ['note_type', 'note_val', 'note_octave']
COLUMNS_RYTHM = ['note_duration', 'silence', 'continuation']
COLUMNS_HARMONIC_RYTHM = ['chord_duration', 'dummy']


def merge_and_update(left, right, on):
    """Merge two sequences by updating when it is possible
    If the right array does not have an entry for the left row, use the previous value of the column

    Parameters
    ----------
    left :
        Sequence that should be updated
    right :
        Sequence acting as the updater
    on :
        str or list, on which columns to perform the left join

    Returns
    -------

    """
    if isinstance(on, str) or isinstance(on, int):
        on = [on]
    new_columns = list(set(right.columns) - set(on))
    final = left.merge(right, on=on, how='left')
    colsx = [col + '_x' for col in new_columns]
    colsy = [col + '_y' for col in new_columns]
    for col in new_columns:
        final[col] = final[col + '_y'].fillna(final[col + '_x'])
    final = final.drop(colsx + colsy, axis=1)
    return final



# Find markov path of chords

def find_words_proba(df, columns, n=4):
    """Find the markov distribution with word of size n

    Parameters
    ----------
    df :
        param columns:
    n :
        return: (Default value = 4)
    columns :
        

    Returns
    -------

    """
    from collections import Counter
    sequence = df[columns]
    # Merge n times
    series = [sequence.add_suffix(f'_{i}').reindex(np.roll(sequence.index, -i)).reset_index(drop=True) for i in
              range(n)]
    df1 = pd.concat(series, axis=1)
    probas = df1.values.tolist()
    nb = len(probas)
    counts = dict(Counter([tuple(m) for m in probas]))
    probas = {key: val / nb for key, val in counts.items()}
    return probas


def find_transitions_proba(df, columns, n=4):
    """

    Parameters
    ----------
    df :
        
    columns :
        
    n :
         (Default value = 4)

    Returns
    -------

    """
    from collections import Counter
    sequence = df[columns]
    assert sequence.isnull().sum().sum() == 0
    sequence = sequence.reset_index(drop=True)
    # Merge n times
    series1 = [sequence.add_suffix(f'_{i}_A').reindex(np.roll(sequence.index, -i)).reset_index(drop=True) for i in
               range(n)]
    series2 = [sequence.add_suffix(f'_{i}_B').reindex(np.roll(sequence.index, -i)).reset_index(drop=True) for i in
               range(n, 2 * n)]
    series = series1 + series2
    df1 = pd.concat(series, axis=1)
    other_columns = [col for col in df1.columns if col.endswith('_B')]
    groups_columns = [col for col in df1.columns if col.endswith('_A')]
    assert df1.isnull().sum().sum() == 0
    all_columns = list(df1.columns)
    df1['count'] = 1
    global_counts = df1.groupby(groups_columns)['count'].count()
    probas = df1.groupby(all_columns)['count'].count()
    probs = probas / global_counts
    # Get dict of dict indexed by A
    groups = probs.groupby(level=list(range(len(columns) * n)))
    res = {}
    for group in groups.groups:
        res[group] = probs.loc[group].to_dict()

    return res


def sample_markov_path(words, transitions, first=None, m=10, n=1):
    """Sample m transitions from a markov path

    Parameters
    ----------
    words :
        
    transitions :
        
    first :
         (Default value = None)
    m :
         (Default value = 10)
    n :
         (Default value = 1)

    Returns
    -------

    """

    def split(a, n):
        """

        Parameters
        ----------
        a :
            
        n :
            

        Returns
        -------

        """

        parts = len(a) // n
        k, m = divmod(len(a), parts)
        return [a[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(parts)]

    def _sample_at_random(D):
        """

        Parameters
        ----------
        D :
            

        Returns
        -------

        """
        return random.choices(list(D.keys()), weights=D.values(), k=1)[0]

    import random
    if first is None:
        previous = _sample_at_random(words[n])
    else:
        previous = first

    N = len(previous) // n
    result = split(previous, N)
    for i in range(m):
        new_sample = _sample_at_random(transitions[n][previous])
        previous = tuple(new_sample)
        N = len(previous) // n
        result += split(new_sample, N)

    return result


def find_distribution(df, columns, n=2):
    """Calculate the words and transitions of words distribution for given columns

    Parameters
    ----------
    df :
        param columns:
    n :
        return: (Default value = 2)
    columns :
        

    Returns
    -------

    """
    words_probs = find_words_proba(df, columns, n=n)
    transitions_probs = find_transitions_proba(df, columns, n=n)

    return words_probs, transitions_probs


def find_chords_distribution(df):
    """Find the chords and chords transitions probabilities of a sequence

    Parameters
    ----------
    df :
        return:

    Returns
    -------

    """
    chords = df.groupby('chord_idx').first()

    N = [1, 2, 3, 4]
    words_result = {}
    transitions_result = {}
    for n in N:
        words_probs, transitions_probs = find_distribution(chords, COLUMNS_CHORDS, n=n)
        words_result[n] = words_probs
        transitions_result[n] = transitions_probs
    return words_result, transitions_result


def find_rythm_distribution(df):
    """Find the rythm and rythm transitions probabilities of a sequence

    Parameters
    ----------
    df :
        return:

    Returns
    -------

    """

    df = df.sort_values(['instrument', 'start'], ascending=True)
    N = [1, 2, 3, 4]
    words_result = {}
    transitions_result = {}
    for n in N:
        words_probs, transitions_probs = find_distribution(df, COLUMNS_RYTHM, n=n)
        words_result[n] = words_probs
        transitions_result[n] = transitions_probs
    return words_result, transitions_result


def find_melody_distribution(df):
    """Find the melody and melody transitions probabilities of a sequence

    Parameters
    ----------
    df :
        return:

    Returns
    -------

    """
    df = df.sort_values(['instrument', 'start'], ascending=True)
    N = [1, 2, 3, 4]
    words_result = {}
    transitions_result = {}
    for n in N:
        words_probs, transitions_probs = find_distribution(df, COLUMNS_MELODY, n=n)
        words_result[n] = words_probs
        transitions_result[n] = transitions_probs
    return words_result, transitions_result


def find_harmonic_rythm_distribution(df):
    """Find the harmonic rythm and harmonic rythm transitions probabilities of a sequence

    Parameters
    ----------
    df :
        return:

    Returns
    -------

    """
    df['chord_duration_max'] = df.groupby('chord_idx')['chord_relative_end'].transform('max')
    df['chord_duration_min'] = df.groupby('chord_idx')['chord_relative_start'].transform('min')
    df['chord_duration'] = df['chord_duration_max'] - df['chord_duration_min']
    chords = df.groupby('chord_idx').first()
    chords['dummy'] = 0
    chords = chords[['chord_duration', 'dummy']]
    N = [1, 2, 3, 4]
    words_result = {}
    transitions_result = {}
    for n in N:
        words_probs, transitions_probs = find_distribution(chords, COLUMNS_HARMONIC_RYTHM, n=n)
        words_result[n] = words_probs
        transitions_result[n] = transitions_probs
    return words_result, transitions_result


def find_distributions(df):
    """

    Parameters
    ----------
    df :
        

    Returns
    -------

    """
    Pc, Tc = find_chords_distribution(df)
    Pr, Tr = find_rythm_distribution(df)
    Pm, Tm = find_melody_distribution(df)
    Ph, Th = find_harmonic_rythm_distribution(df)
    
    return Pc, Tc, Pr, Tr, Pm, Tm, Ph, Th

def sample_score(Pc, Tc, Pr, Tr, Pm, Tm, Ph, Th, duration=10, n_c=1, n_h=1, n_r=1, n_m=4):
    """

    Parameters
    ----------
    Pc :
        
    Tc :
        
    Pr :
        
    Tr :
        
    Pm :
        
    Tm :
        
    Ph :
        
    Th :
        
    duration :
         (Default value = 10)
    n_c :
         (Default value = 1)
    n_h :
         (Default value = 1)
    n_r :
         (Default value = 1)
    n_m :
         (Default value = 4)

    Returns
    -------

    """
    import pandas as pd
    def _sample_with_duration_constraint(P, T, idx_duration, duration, n=1):
        """

        Parameters
        ----------
        P :
            
        T :
            
        idx_duration :
            
        duration :
            
        n :
             (Default value = 1)

        Returns
        -------

        """
        import numpy as np
        # Sample harmonic rythm
        m = 10
        while True:
            path = sample_markov_path(P, T, first=None, m=m, n=n)
            array_duration = np.asarray([p[idx_duration] for p in path])
            sample_duration = np.sum(array_duration)
            if sample_duration > duration:
                cumsum = np.cumsum(array_duration)
                idx_break = np.argmax(cumsum >= duration)
                to_remove = cumsum[idx_break] - duration
                path[idx_break] = list(path[idx_break])
                path[idx_break][idx_duration] = path[idx_break][idx_duration] - to_remove
                path[idx_break] = tuple(path[idx_break])
                path = path[:idx_break + 1]
                break
            elif sample_duration == duration:
                break
            else:
                m = 2 * m
        return path

    path_h = _sample_with_duration_constraint(Ph, Th, 0, duration, n=n_h)
    path_c = sample_markov_path(Pc, Tc, first=None, m=len(path_h), n=n_c)
    path_r = _sample_with_duration_constraint(Pr, Tr, 0, duration, n=n_r)
    path_m = sample_markov_path(Pm, Tm, first=None, m=len(path_r), n=n_m)

    path_c = path_c[:len(path_h)]
    path_m = path_m[:len(path_r)]
    assert len(path_h) == len(path_c), (len(path_h), len(path_c))
    assert len(path_r) == len(path_m), (len(path_r), len(path_m))
    # Create melody serie
    melody_df = pd.DataFrame(path_m, columns=COLUMNS_MELODY)
    rythm_df = pd.DataFrame(path_r, columns=COLUMNS_RYTHM)
    harmonic_rythm_df = pd.DataFrame(path_h, columns=COLUMNS_HARMONIC_RYTHM)
    chords_df = pd.DataFrame(path_c, columns=COLUMNS_CHORDS)

    notes = pd.concat([melody_df, rythm_df], axis=1)
    notes.loc[notes['silence'] == True, 'note_type'] = 'r'
    notes.loc[notes['continuation'] == True, 'note_type'] = 'l'
    notes.loc[notes['silence'] == True, 'note_val'] = 0
    notes.loc[notes['continuation'] == True, 'note_val'] = 0
    notes.loc[notes['silence'] == True, 'note_octave'] = 0
    notes.loc[notes['continuation'] == True, 'note_octave'] = 0
    notes['end'] = notes['note_duration'].cumsum()
    notes['start'] = notes['end'].shift(1).fillna(0)

    chords = pd.concat([harmonic_rythm_df, chords_df], axis=1)
    chords['chord_idx'] = np.arange(0, len(chords))
    chords['chord_end'] = chords['chord_duration'].cumsum()
    chords['chord_start'] = chords['chord_end'].shift(1).fillna(0)
    # Create chord serie
    merged = notes.merge(chords, how='cross')
    # Filter null intersection
    null_intersection = (merged['end'] <= merged['chord_start']) | (merged['start'] >= merged['chord_end'])
    merged = merged[~null_intersection]
    # Merge notes and chords and then fix durations
    merged['start_removal'] = (merged['chord_start'] - merged['start']).abs() * (
                merged['start'] < merged['chord_start'])
    merged['end_removal'] = (merged['end'] - merged['chord_end']).abs() * (merged['chord_end'] < merged['end'])
    merged['start'] += merged['start_removal']
    merged['end'] -= merged['end_removal']
    merged.loc[merged['start_removal'] > 0, 'note_type'] = 'l'
    # Add missing columns to be parsed by engine
    merged['tonality_octave'] = 0
    merged['chord_extension'] = 5
    merged['instrument'] = 'piano__0'
    merged['note_amp'] = 100
    merged['note_duration'] = merged['end'] - merged['start']
    merged = merged.sort_values('start')

    return merged



from musiclang.transform import TransformPipeline, ConcatPipeline, NoteTransformer
from musiclang.library import *
from musiclang.transform import Mask


class Transpose(NoteTransformer):

    def __init__(self, n):
        self.n = n

    def action(self, note, chord=None, instrument=None, **kwargs):

        return note & self.n

def test_transform_pipeline_without_mask():
    transposer = Transpose(2)
    pipeline = TransformPipeline([(
        'transpose', transposer
    )])

    score = (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4])
    expected_result = (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s2, s4, s6])
    new_score = pipeline(score)
    assert new_score == expected_result

def test_transform_pipeline_without_mask_several():
    transposer = Transpose(2)
    pipeline = TransformPipeline([
        ('transpose', transposer),
        ('transpose', transposer),
    ])

    score = (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4])
    expected_result = score & 4
    new_score = pipeline(score)
    assert new_score == expected_result

def test_transform_pipeline_with_mask():
    transposer = Transpose(2)
    mask = Mask.Chord() > Mask.Has('ok')

    pipeline = TransformPipeline([(
        'transpose', transposer, mask
    )])

    score = (I % I.M)(piano=[s0, s2, s4]).add_tag('ok') + (V % I.M)(piano=[s0, s2, s4])
    expected_result = (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s0, s2, s4])
    new_score = pipeline(score)
    assert new_score == expected_result


def test_concat_pipeline_without_mask():
    transposer = Transpose(2)
    pipeline = ConcatPipeline([
        ('transpose', transposer),
        ('transpose', transposer),
    ])

    score = (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4])
    expected_result = score + (score & 2) + ((score & 2) + (score & 4))
    new_score = pipeline(score)
    assert new_score == expected_result


def test_concat_pipeline_without_mask_several():
    transposer = Transpose(2)
    pipeline = ConcatPipeline([(
        'transpose', transposer
    )])

    score = (I % I.M)(piano=[s0, s2, s4]) + (V % I.M)(piano=[s0, s2, s4])
    expected_result = score + (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s2, s4, s6])
    new_score = pipeline(score)
    assert new_score == expected_result


def test_concat_pipeline_with_mask():
    transposer = Transpose(2)
    mask = Mask.Chord() > Mask.Has('ok')

    pipeline = ConcatPipeline([(
        'transpose', transposer, mask
    )])

    score = (I % I.M)(piano=[s0, s2, s4]).add_tag('ok') + (V % I.M)(piano=[s0, s2, s4])
    expected_result = score + (I % I.M)(piano=[s2, s4, s6]) + (V % I.M)(piano=[s0, s2, s4])
    new_score = pipeline(score)
    assert new_score == expected_result



"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

class Item(object):
    """ """
    def __init__(self, name, start, end, vel=0, pitch=0, track=0, channel=0, voice=0, value=''):
        self.name = name
        self.start = start  # start step
        self.end = end  # end step
        self.vel = vel
        self.pitch = pitch
        self.track = track
        self.channel = channel
        self.voice = voice
        self.value = value


    @classmethod
    def frommatrix(cls, matrix):
        """

        Parameters
        ----------
        matrix :
            

        Returns
        -------

        """
        return [Item('', start, end, vel, pitch, track, voice, channel) for start, end, vel, pitch, track, voice, channel in matrix]

    def __repr__(self):
        return f'Item(name={self.name:>10s}, start={self.start}, end={self.end}, ' \
               f'vel={self.vel}, pitch={self.pitch}, track={self.track}, channel={self.channel}, voice={self.voice},' \
               f'value={self.value})\n'

    def __eq__(self, other):
        return self.name == other.name and self.start == other.start and \
               self.pitch == other.pitch and self.track == other.track

    def array(self):
        """ """
        return [self.start, self.end, self.vel, self.pitch, self.track, self.channel, self.voice]



def quantize_notes_raw(notes, limit_denominator=8):
    """Quantize a sequence of notes limiting the denominator to "limit_denominator" parameter

    Parameters
    ----------
    notes :
        List[Item] : sequence of notes
    limit_denominator :
         (Default value = 8)

    Returns
    -------
    type
        new_notes : List[Item], quantized notes

    """
    from fractions import Fraction as frac
    new_notes = []
    for note in notes:
        note.start = note.start
        note.end = note.end
        note.start = frac(note.start).limit_denominator(limit_denominator)
        note.end = frac(note.end).limit_denominator(limit_denominator)
        new_notes.append(note)

    return new_notes


def convert_to_items(notes):
    """Convert an array of notes into an array of Item objects that represents notes
    and quantize them

    Parameters
    ----------
    notes :
        param bar_duration:
    offset :
        return:

    Returns
    -------

    """
    # Quantize
    items = []
    for note in notes:
        start = note[0]
        end = start + note[2]
        vel = int(note[3])
        pitch = int(note[1])
        track = int(note[4])
        channel = int(note[5])
        voice = int(note[6])
        items.append(Item("name", start, end, vel=vel, pitch=pitch, track=track, channel=channel, voice=voice))

    # Quantize notes in integer
    items = quantize_notes_raw(items)
    return items


from hashlib import sha256
from struct import unpack
import numpy as np

def bytes_to_float(b):
    return float(unpack('L', sha256(b).digest()[:8])[0]) / 2**64


def random_permutation(chars, array, seed="", root=""):
    """
    Generate a permutation of size n following a bytes encoding
    if chars,seed,root,proba is the same it always generates the same output
    Algorithm : Fisher Yates shuffle  : https://en.wikipedia.org/wiki/Random_permutation#:~:text=A%20simple%20algorithm%20to%20generate,has%20index%200%2C%20and%20the
    :param array: array to shuffle
    :param seed:
    :param root:
    :return:
    """
    n = len(array)
    for i in range(n - 1):
        j = i + random_int(chars + str(i), low=0, high=n - i, seed=seed, root=root)
        z = array[i]
        array[i] = array[j]
        array[j] = z

    return array


def random_boolean(chars, seed="", root="", proba=0.2):
    """
    Generate a random boolean following a bytes encoding
    if chars,seed,root,proba is the same it always generates the same output
    :param seed:
    :param root:
    :param chars:
    :param proba:
    :return:
    """
    text = f'{chars}{seed}{root}'
    text = bytes(text, encoding='utf-8')
    return bytes_to_float(text) < proba


def random_int(chars, low, high, seed="", root=""):
    """
    Generate an integer between low and high following a bytes encoding
    if chars,seed,root,proba is the same it always generates the same output
    :param chars:
    :param low:
    :param high:
    :param seed:
    :param root:
    :return:
    """
    return int(random_float(chars, low, high, seed=seed, root=root))


def random_float(chars, low, high, seed="", root=""):
    """
    Generate a float between low and high following a bytes encoding
    if chars,seed,root,proba is the same it always generates the same output
    :param chars:
    :param low:
    :param high:
    :param seed:
    :param root:
    :return:
    """
    text = f'{chars}{seed}{root}'
    text = bytes(text, encoding='utf-8')
    return (high - low) * bytes_to_float(text) + low

def init_seed(seed):
    if seed is None:
        return np.random.randint(0, 2 ** 31)
    else:
        return seed


def random_choice(array, chars, seed="", root=""):
    return array[random_int(chars, 0, len(array), seed=seed, root=root)]


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

from .constants import *

class Tonality:
    """Represents a tonality.

    It can be applied with the ``%`` operator to a chord to modulate it on another tonality
    It encapsulates a degree, a mode, and an octave


    """
    def __init__(self, degree, mode="M", octave=0, tags=None):
        """
        Initialize the tonality.
        Usually you will use musiclang.library to instantiate the tonalities

        Parameters
        ----------

        degree: Degree
                Degree is absolute degree (between 0 and 12). For example 6 is f#, 11 is b

        mode: str in ["M", "m", "mm", "dorian", "phrygian", "lydian", "mixolydian", "aeolian", "locrian"]

        octave: int, default=0
                Octave of the tonality, by default it will be 0 which represents the 5th octave on a piano.
        """
        self.degree = degree
        self.mode = mode
        self.octave = octave
        self.tags = set(tags) if tags is not None else set()


    @classmethod
    def from_str(cls, text):
        tab = 'CDEFGAB'
        notes = [0, 2, 4, 5, 7, 9, 11]
        note = text.replace(':', '').replace('#', '').replace('b', '').replace('-', '')
        tone = notes[tab.index(note.upper())]
        mode = 'M' if note.upper() == note else 'm'

        tone += text.count('#')
        tone += text.count('s')
        tone -= text.count('b')
        tone -= text.count('-')

        return cls(tone, mode)

    def __hash__(self):
        return hash(self.__repr__())


    def has_tag(self, tag):
        """
        Check if the tag exists for this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        tonality: Tonality
        """
        return tag in self.tags

    def add_tag(self, tag):
        """
        Add a tag to this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        tonality: Tonality
        """
        cp = self.copy()
        cp.tags.add(tag)
        return cp

    def remove_tag(self, tag):
        """
        Remove a tag from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        tonality: Tonality
        """
        cp = self.copy()
        cp.tags.remove(tag)
        return cp

    def add_tags(self, tags):
        """
        Add several tags to the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        tonality: Tonality

        """
        cp = self.copy()
        cp.tags = cp.tags.union(set(tags))
        return cp

    def remove_tags(self, tags):
        """
        Remove several tags from the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]

        Returns
        -------
        tonality: Tonality


        """
        cp = self.copy()
        cp.tags = cp.tags - set(tags)
        return cp

    def clear_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        tonality: Tonality
        """
        cp = self.copy()
        cp.tags = set()
        return cp

    def change_mode(self, mode):
        """

        Parameters
        ----------
        mode :
            

        Returns
        -------

        """
        new_tonality = self.copy()
        new_tonality.mode = mode
        return new_tonality


    @property
    def scale_set(self):
        """ """
        return frozenset({s % 12 for s in self.scale_pitches})

    @property
    def scale_pitches(self):
        """
        Get the scale absolute pitches (integers)

        See Also
        --------
        :func:`~Chord.scale_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> I.M.scale_pitches
        [0, 2, 4, 5, 7, 9, 11]

        """
        abs_degree = self.abs_degree
        mode = self.mode
        pitch_scale = [n + abs_degree for n in SCALES[mode]]
        return pitch_scale

    def _eq(self, other):
        """

        Parameters
        ----------
        other :
            

        Returns
        -------

        """
        return self.degree == other.degree and self.mode == other.mode and self.octave == other.octave

    def __eq__(self, other):
        if not isinstance(other, Tonality):
            return False
        # To get equalities between enharmonies we need to reformulate chord adding neutral modulation
        return (Tonality(0) + self )._eq(Tonality(0) + other )

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def copy(self):
        """
        Returns a copy of the tonality
        """
        return Tonality(self.degree, self.mode, self.octave, tags=set(self.tags))

    def __radd__(self, other):
        if other is None:
            return self
        else:
            raise Exception('Not valid addition of tonality')

    def add(self, other):
        """
        Add another tonality to this one
        - The degree is added (modulo 12 with octave)
        - The mode is the mode of the right tonality

        Parameters
        ----------
        other : Tonality
            

        Returns
        -------

        """
        new_abs_degree = self.degree + other.degree
        new_octave = self.octave + other.octave
        delta_octave = new_abs_degree // 12
        new_degree = new_abs_degree % 12
        new_mode = other.mode
        return Tonality(degree=new_degree, mode=new_mode, octave=new_octave + delta_octave, tags=self.tags.union(other.tags))

    @property
    def abs_degree(self):
        """ """
        return self.degree + 12 * self.octave

    def __add__(self, other):
        """
        Add another tonality
        :param other:
        :return:
        """
        if other is None:
            return self
        else:
            return self.add(other)

    def __sub__(self, other):
        """
        Find the relative tonality to go from other to self
        Parameters
        ----------
        other: Tonality

        Returns
        -------
        tonality: Tonality

        """
        new_abs_degree = self.degree - other.degree
        new_octave = self.octave - other.octave
        delta_octave = new_abs_degree // 12
        new_degree = new_abs_degree % 12
        new_mode = self.mode
        return Tonality(degree=new_degree, mode=new_mode, octave=new_octave + delta_octave,
                        tags=self.tags.union(other.tags))

    @property
    def b(self):
        """ """
        tone = self.copy()
        tone.degree -= 1
        if tone.degree == -1:
            tone.degree = 11
            tone.octave -= 1
        return tone

    @property
    def s(self):
        """ """
        tone = self.copy()
        tone.degree += 1
        if tone.degree == 12:
            tone.degree = 0
            tone.octave += 1
        return tone

    @property
    def m(self):
        """ """
        tone = self.copy()
        tone.mode = "m"
        return tone

    @property
    def M(self):
        """ """
        tone = self.copy()
        tone.mode = "M"
        return tone

    @property
    def mm(self):
        """ """
        tone = self.copy()
        tone.mode = "mm"
        return tone


    @property
    def dorian(self):
        """ """
        tone = self.copy()
        tone.mode = "dorian"
        return tone

    @property
    def phrygian(self):
        """ """
        tone = self.copy()
        tone.mode = "phrygian"
        return tone

    @property
    def lydian(self):
        """ """
        tone = self.copy()
        tone.mode = "lydian"
        return tone

    @property
    def mixolydian(self):
        """ """
        tone = self.copy()
        tone.mode = "mixolydian"
        return tone

    @property
    def aeolian(self):
        """ """
        tone = self.copy()
        tone.mode = "aeolian"
        return tone

    @property
    def locrian(self):
        """ """
        tone = self.copy()
        tone.mode = "locrian"
        return tone


    def d(self):
        """ """
        return self.o(-1)

    def o(self, octave):
        """
        Octave the tonality



        Parameters
        ----------
        octave : int
            

        Returns
        -------
        tonality: Tonality

        """
        tonality = self.copy()
        tonality.octave += octave
        return tonality

    def degree_to_str(self):
        """ """
        return DEGREE_TO_STR[self.degree]

    def __repr__(self):
        return self.to_code()

    def to_code(self):
        """
        Represent the tonality in valid Python code

        Returns
        -------
        tonality_code: str

        """
        result = f"{self.degree_to_str()}.{self.mode}"
        if self.octave != 0:
            result += f".o({self.octave})"

        return result

    def __call__(self, *notes, **kwargs):
        from .custom_chord import CustomChord
        return CustomChord(notes, tonality=self, **kwargs)


    def to_absolute_romantext(self):
        L = ['c', 'db', 'd', 'eb', 'e', 'f', 'f#', 'g', 'ab', 'a', 'bb', 'b']
        return L[self.degree].capitalize() if self.mode == 'M' else L[self.degree]

    def to_romantext(self, base_tonality):
        base_tone = base_tonality.degree
        tone = (self.degree - base_tone) % 12
        base_mode = base_tonality.mode
        mode = self.mode

        L = {
            ('m', 'm'): ['i', 'bii', 'ii', 'iii', '#iii', 'iv', '#iv', 'v', 'bvi', 'vi', 'bvii', 'vii'],
            ('M', 'm'): ['I', 'bII', 'II', 'III', '#III', 'IV', '#IV', 'V', 'VI', '#VI', 'VII', '#VII'],
            ('m', 'M'): ['i', 'bii', 'ii', 'biii', 'iii', 'iv', '#iv', 'v', 'bvi', 'vi', 'bvii', 'vii'],
            ('M', 'M'): ['I', 'bII', 'II', 'bIII', 'III', 'IV', '#IV', 'V', 'bVI', 'VI', 'bVII', 'vii'],
        }
        return L[(mode, base_mode)][tone]


"""Tonal representations used as inputs to the network."""

import numpy as np
import re

from .cache import (
    TransposePitch,
    m21Pitch,
    m21IntervalStr,
)
from .feature_representation import (
    INTERVALCLASSES,
    NOTEDURATIONS,
    NOTENAMES,
    PITCHCLASSES,
    SPELLINGS,
    FeatureRepresentation,
    FeatureRepresentationTI,
)


class MeasureOnset7(FeatureRepresentationTI):
    """ """
    features = len(NOTEDURATIONS)
    pattern = [list(reversed(f"{x:06b}0")) for x in range(64)]
    pattern = [[int(n) for n in arr] for arr in pattern]
    pattern[0][0] = 1

    def run(self, transposition=None):
        """

        Parameters
        ----------
        transposition :
             (Default value = None)

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        prev_measure = -1
        idx = 0
        for frame, measure in enumerate(self.df.s_measure):
            if measure != prev_measure:
                idx = 0
                prev_measure = measure
            pattern = self.pattern[idx]
            array[frame] = pattern
            idx = min(idx + 1, len(self.pattern) - 1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            measureOnset = [NOTEDURATIONS[x] for x in np.nonzero(manyhot)[0]]
            ret.append(measureOnset)
        return ret


class NoteOnset7(MeasureOnset7):
    """ """
    def run(self, transposition=None):
        """

        Parameters
        ----------
        transposition :
             (Default value = None)

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        idx = 0
        for frame, onset in enumerate(self.df.s_isOnset):
            if sum(onset) > 0:
                idx = 0
            pattern = self.pattern[idx]
            array[frame] = pattern
            idx = min(idx + 1, len(self.pattern) - 1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            noteOnset = [NOTEDURATIONS[x] for x in np.nonzero(manyhot)[0]]
            ret.append(noteOnset)
        return ret


class MeasureNoteOnset14(FeatureRepresentationTI):
    """ """
    features = MeasureOnset7.features + NoteOnset7.features
    pattern = MeasureOnset7.pattern

    def run(self, transposition=None):
        """

        Parameters
        ----------
        transposition :
             (Default value = None)

        Returns
        -------

        """
        self.measure7 = MeasureOnset7(self.df).run(transposition)
        self.note7 = NoteOnset7(self.df).run(transposition)
        array = np.concatenate((self.measure7, self.note7), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        measure7 = MeasureOnset7.decode(array[:, : MeasureOnset7.features])
        note7 = NoteOnset7.decode(array[:, MeasureOnset7.features :])
        return [(tuple(mm), tuple(n)) for mm, n in zip(measure7, note7)]


class Bass12(FeatureRepresentation):
    """ """
    features = len(PITCHCLASSES)

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, notes in enumerate(self.df.s_notes):
            bass = notes[0]
            transposed = TransposePitch(bass, transposition)
            pitchObj = m21Pitch(transposed)
            pitchClass = pitchObj.pitchClass
            array[frame, pitchClass] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            bassPitchClass = np.argmax(manyhot)
            ret.append(bassPitchClass)
        return ret


class Bass7(FeatureRepresentation):
    """ """
    features = len(NOTENAMES)

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, notes in enumerate(self.df.s_notes):
            bass = notes[0]
            transposed = TransposePitch(bass, transposition)
            pitchObj = m21Pitch(transposed)
            pitchLetter = pitchObj.step
            pitchLetterIndex = NOTENAMES.index(pitchLetter)
            array[frame, pitchLetterIndex] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            bassPitchName = NOTENAMES[np.argmax(manyhot)]
            ret.append(bassPitchName)
        return ret


class Bass19(FeatureRepresentation):
    """ """
    features = Bass12.features + Bass7.features

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        self.letter = Bass7(self.df).run(transposition)
        self.pc = Bass12(self.df).run(transposition)
        array = np.concatenate((self.letter, self.pc), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        letters = Bass7.decode(array[:, : Bass7.features])
        pcs = Bass12.decode(array[:, Bass7.features :])
        return [(l, pc) for l, pc in zip(letters, pcs)]


class Chromagram12(FeatureRepresentation):
    """ """
    features = len(PITCHCLASSES)

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, notes in enumerate(self.df.s_notes):
            for note in notes:
                transposedNote = TransposePitch(note, transposition)
                pitchObj = m21Pitch(transposedNote)
                pitchClass = pitchObj.pitchClass
                array[frame, pitchClass] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            chromagramPitchClasses = np.nonzero(manyhot)[0].tolist()
            ret.append(tuple(chromagramPitchClasses))
        return ret


class Chromagram7(FeatureRepresentation):
    """ """
    features = len(NOTENAMES)

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, notes in enumerate(self.df.s_notes):
            for note in notes:
                transposedNote = TransposePitch(note, transposition)
                pitchObj = m21Pitch(transposedNote)
                pitchLetter = pitchObj.step
                pitchLetterIndex = NOTENAMES.index(pitchLetter)
                array[frame, pitchLetterIndex] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            pitchNames = [NOTENAMES[x] for x in np.nonzero(manyhot)[0]]
            ret.append(tuple(pitchNames))
        return ret


class Chromagram19(FeatureRepresentation):
    """ """
    features = Chromagram12.features + Chromagram7.features

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        self.letter = Chromagram7(self.df).run(transposition)
        self.pc = Chromagram12(self.df).run(transposition)
        array = np.concatenate((self.letter, self.pc), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        letters = Chromagram7.decode(array[:, : Chromagram7.features])
        pcs = Chromagram12.decode(array[:, Chromagram7.features :])
        return [(l, pc) for l, pc in zip(letters, pcs)]


class Intervals19(FeatureRepresentationTI):
    """ """
    features = len(NOTENAMES) + len(PITCHCLASSES)

    def run(self):
        """ """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, intervals in enumerate(self.df.s_intervals):
            for interval in intervals:
                intervalObj = m21IntervalStr(interval)
                chromatic = intervalObj.chromatic.mod12
                genericClass = intervalObj.generic.simpleUndirected - 1
                array[frame, genericClass] = 1
                array[frame, chromatic + len(NOTENAMES)] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            generics = (np.nonzero(manyhot[:7])[0] + 1).tolist()
            chromatics = np.nonzero(manyhot[7:])[0].tolist()
            ret.append((tuple(generics), tuple(chromatics)))
        return ret


class Intervals39(FeatureRepresentationTI):
    """ """
    features = len(INTERVALCLASSES)

    def run(self):
        """ """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, intervals in enumerate(self.df.s_intervals):
            for interval in intervals:
                intervalIndex = INTERVALCLASSES.index(interval)
                array[frame, intervalIndex] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != cls.features:
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            intervals = [INTERVALCLASSES[x] for x in np.nonzero(manyhot)[0]]
            ret.append(tuple(intervals))
        return ret


class Bass35(FeatureRepresentation):
    """ """
    features = len(SPELLINGS)

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, notes in enumerate(self.df.s_notes):
            bass = re.sub(r"\d", "", notes[0])
            transposed = TransposePitch(bass, transposition)
            if transposed in SPELLINGS:
                spellingIndex = SPELLINGS.index(transposed)
                array[frame, spellingIndex] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != len(SPELLINGS):
            raise IndexError("Strange array shape.")
        return [SPELLINGS[np.argmax(onehot)] for onehot in array]


class Chromagram35(FeatureRepresentation):
    """ """
    features = len(SPELLINGS)

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        array = np.zeros(self.shape, dtype=self.dtype)
        for frame, notes in enumerate(self.df.s_notes):
            notes = [re.sub(r"\d", "", n) for n in notes]
            for note in notes:
                transposed = TransposePitch(note, transposition)
                if transposed in SPELLINGS:
                    spellingIndex = SPELLINGS.index(transposed)
                    array[frame, spellingIndex] = 1
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        if len(array.shape) != 2 or array.shape[1] != len(SPELLINGS):
            raise IndexError("Strange array shape.")
        ret = []
        for manyhot in array:
            notes = [SPELLINGS[x] for x in np.nonzero(manyhot)[0]]
            ret.append(tuple(notes))
        return ret


class BassChromagram70(FeatureRepresentation):
    """ """
    features = Bass35.features + Chromagram35.features

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        self.bass35 = Bass35(self.df).run(transposition)
        self.chromagram35 = Chromagram35(self.df).run(transposition)
        array = np.concatenate((self.bass35, self.chromagram35), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        bass35 = Bass35.decode(array[:, : Bass35.features])
        chromagram35 = Chromagram35.decode(array[:, Bass35.features :])
        return [(b, ch) for b, ch in zip(bass35, chromagram35)]


class BassChromagram38(FeatureRepresentation):
    """ """
    features = Bass19.features + Chromagram19.features

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        # super().__init__(df)
        self.bass19 = Bass19(self.df).run(transposition)
        self.chromagram19 = Chromagram19(self.df).run(transposition)
        array = np.concatenate((self.bass19, self.chromagram19), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        bass19 = Bass19.decode(array[:, : Bass19.features])
        chromagram19 = Chromagram19.decode(array[:, Bass19.features :])
        return [(b[0], b[1], c[0], c[1]) for b, c in zip(bass19, chromagram19)]


class BassIntervals58(FeatureRepresentation):
    """ """
    features = Bass19.features + Intervals39.features

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        self.bass19 = Bass19(self.df).run(transposition)
        self.intervals39 = Intervals39(self.df).run()
        array = np.concatenate((self.bass19, self.intervals39), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        bass19 = Bass19.decode(array[:, : Bass19.features])
        intervals39 = Intervals39.decode(array[:, Bass19.features :])
        return [(b[0], b[1], i) for b, i in zip(bass19, intervals39)]


class BassChromagramIntervals77(FeatureRepresentation):
    """ """
    features = BassChromagram38.features + Intervals39.features

    def run(self, transposition="P1"):
        """

        Parameters
        ----------
        transposition :
             (Default value = "P1")

        Returns
        -------

        """
        self.bassChroma38 = BassChromagram38(self.df).run(transposition)
        self.intervals39 = Intervals39(self.df).run()
        array = np.concatenate((self.bassChroma38, self.intervals39), axis=1)
        return array

    @classmethod
    def decode(cls, array):
        """

        Parameters
        ----------
        array :
            

        Returns
        -------

        """
        bassChroma38 = BassChromagram38.decode(
            array[:, : BassChromagram38.features]
        )
        intervals39 = Intervals39.decode(array[:, BassChromagram38.features :])
        return [
            (bc[0], bc[1], bc[2], bc[3], i)
            for bc, i in zip(bassChroma38, intervals39)
        ]


available_representations = {
    "Bass7": Bass7,
    "Bass12": Bass12,
    "Bass19": Bass19,
    "Bass35": Bass35,
    "Chromagram7": Chromagram7,
    "Chromagram12": Chromagram12,
    "Chromagram19": Chromagram19,
    "Chromagram35": Chromagram35,
    "MeasureOnset7": MeasureOnset7,
    "NoteOnset7": NoteOnset7,
    "MeasureNoteOnset14": MeasureNoteOnset14,
    "Intervals39": Intervals39,
    "Intervals19": Intervals19,
    "BassChromagram38": BassChromagram38,
    "BassChromagram70": BassChromagram70,
    "BassIntervals58": BassIntervals58,
    "BassChromagramIntervals77": BassChromagramIntervals77,
}
